/*
 * File:   GraphWindow.java
 * Author: Conrad Rider
 * Email:  c.rider@eviltree.co.uk
 * Date:   04-03-05
 */
 
 package gui.graph;
 
import gui.*;
import gui.view.*;
import gui.context.*;
import sim.Simulation;
import sim.livestock.*;
import main.PastureSim;
import utils.GraphData;

import java.awt.*;
import java.awt.Color;
import java.awt.Container;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.io.*;
import javax.swing.*;
import javax.swing.SpringLayout;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JTextField;
import javax.swing.JButton;

/** The abstract class representing a graphing windodw */
public abstract class GraphWindow extends JFrame implements ActionListener{

	public static final int WIDTH = 600;
	public static final int HEIGHT = 350;
	public static final int CONTEXT_HEIGHT = 45;
	
	protected double scaleLen;
	protected double scaleEnd;
	
	protected JTextField periodVal;
	protected JTextField graphToVal;
	protected JButton update;
	protected JCheckBox[] plots;

		
	protected UserInterface ui; // A reference to the user interface
	protected Simulation sim; // A reference to the simulation
	
	
	
	protected Container mainPanel;
	protected SpringLayout layout;
	protected GraphPanel viewPanel;
	
	
		
	public GraphWindow(UserInterface ui, Simulation sim, String title){
		super(title);
		this.ui = ui;
		this.sim = sim;
		
		scaleLen = 0.0;
		scaleEnd = 0.0;
		
		setResizable(true);
		setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		setIconImage(UserInterface.iconImage);

		WindowListener closeAction = new WindowListener(){
			public void windowActivated(WindowEvent e){}
			public void windowClosed(WindowEvent e){}
			public void windowClosing(WindowEvent e){ close(); }
			public void windowDeactivated(WindowEvent e){}
			public void windowDeiconified(WindowEvent e){}
			public void windowIconified(WindowEvent e){}
			public void windowOpened(WindowEvent e){}
		}; addWindowListener(closeAction);
		
		//Schedule a job for the event-dispatching thread:
		//creating and showing this application's GUI.
		(new Thread(){ public void run(){ generate(); } }).start();
		
	}
	
	
	/** Generate the gui */
	protected void generate(){
		mainPanel = getContentPane();
		layout = new SpringLayout();
		mainPanel.setLayout(layout);

		viewPanel = new GraphPanel(WIDTH, HEIGHT); mainPanel.add(viewPanel);
		layout.putConstraint(SpringLayout.NORTH, viewPanel,  5, SpringLayout.NORTH, mainPanel);
		layout.putConstraint(SpringLayout.WEST,  viewPanel,  5, SpringLayout.WEST,  mainPanel);
		layout.putConstraint(SpringLayout.SOUTH, mainPanel, 55, SpringLayout.SOUTH, viewPanel);
		
		double maxWidth = 0.0;
		if(plots != null && plots.length > 0){
			mainPanel.add(plots[0]);
			maxWidth = plots[0].getPreferredSize().getWidth();;
			layout.putConstraint(SpringLayout.NORTH, plots[0], 10, SpringLayout.NORTH, viewPanel);
			layout.putConstraint(SpringLayout.WEST, plots[0],  10, SpringLayout.EAST, viewPanel);
		}else{
			layout.putConstraint(SpringLayout.EAST, mainPanel, 0, SpringLayout.EAST, viewPanel);
		}
		for(int i = 1; i < plots.length; i++){
			if(plots[i].getPreferredSize().getWidth() > maxWidth) maxWidth = plots[i].getPreferredSize().getWidth();
			mainPanel.add(plots[i]);
			layout.putConstraint(SpringLayout.NORTH, plots[i], 10, SpringLayout.SOUTH, plots[i-1]);
			layout.putConstraint(SpringLayout.WEST,  plots[i],  0, SpringLayout.WEST, plots[i-1]);
		}
		layout.putConstraint(SpringLayout.EAST,  mainPanel, (int)maxWidth+15, SpringLayout.EAST,  viewPanel);

				
		JLabel period  = new JLabel("Period [days, 0 for all]");    mainPanel.add(period);
		periodVal      = new JTextField(""+scaleLen, 12);      mainPanel.add(periodVal);
		JLabel graphTo = new JLabel("Up To [days ago]");       mainPanel.add(graphTo);
		graphToVal     = new JTextField(""+scaleEnd, 12);      mainPanel.add(graphToVal);
		update         = new JButton("Update Axis");           mainPanel.add(update); update.addActionListener(this);
		
		layout.putConstraint(SpringLayout.NORTH, period,    7, SpringLayout.SOUTH, viewPanel);
		layout.putConstraint(SpringLayout.NORTH, graphTo,   7, SpringLayout.SOUTH, viewPanel);
		
		layout.putConstraint(SpringLayout.NORTH, periodVal,  5, SpringLayout.SOUTH, period);
		layout.putConstraint(SpringLayout.NORTH, graphToVal, 3, SpringLayout.SOUTH, graphTo);

		layout.putConstraint(SpringLayout.WEST, period,     10, SpringLayout.WEST, this);
		layout.putConstraint(SpringLayout.WEST, periodVal,   0, SpringLayout.WEST, period);
		layout.putConstraint(SpringLayout.WEST, graphTo,    40, SpringLayout.EAST, period);
		layout.putConstraint(SpringLayout.WEST, graphToVal,  0, SpringLayout.WEST, graphTo);
		
		layout.putConstraint(SpringLayout.EAST, update,      0, SpringLayout.EAST, viewPanel);
		layout.putConstraint(SpringLayout.SOUTH, update,    -2, SpringLayout.SOUTH, periodVal);
		
		setLocationRelativeTo(ui);
		pack();
		setVisible(true);
		
	}
	
	
	/** Updates the graph */
	public final void update(){ viewPanel.repaint(); }
	
	
	/** Plots the data */
	public abstract void drawGraph(Graphics g);
	
	
	private void close(){ ui.notifyClosed(this); dispose(); }

	public void enableButtons(){
		update.setEnabled(true);
	}
	
	public void disableButtons(){
		update.setEnabled(false);
	}
	
	public void actionPerformed(ActionEvent e){
		if(e.getSource() == update){
			try{
				scaleLen = Double.parseDouble(periodVal.getText());
				periodVal.setText(""+scaleLen);
			}catch(NumberFormatException ex){
				periodVal.setText(""+scaleLen);
				gui.Popup.error("Period must be a valid number.");
			}
			try{
				scaleEnd = Double.parseDouble(graphToVal.getText());
				graphToVal.setText(""+scaleEnd);
			}catch(NumberFormatException ex){
				graphToVal.setText(""+scaleEnd);
				gui.Popup.error("'Up To' must be a valid number.");
			}
			update();
		}
		else if(plots != null && plots.length > 0){
			/*
			for(int i = 0; i < plots.length; i++){
				if(e.getSource() == plots[i]){
					if(plots[i].isSelected())
					else 
				}
			}
			*/
			update();
		}
	}

	/** The panel for drawing graphs */
	class GraphPanel extends JPanel{


		public GraphPanel(int w, int h){
			setBackground(Color.WHITE);
			setBorder(BorderFactory.createLoweredBevelBorder());

			setPreferredSize(new Dimension(w, h));
		}

		/** Draws the graph */
		public void paintComponent(Graphics g){ super.paintComponent(g); drawGraph(g); drawAxis(g); }
	}
	
	/** Plots a minutely data set */
	protected void plotMinutes(GraphData data, Graphics g){
		
		int end = (int)(sim.minute()/1440.0-scaleEnd);
		int len = (scaleLen <= 0.0)? (int)(sim.minute()/1440.0-scaleEnd) : (int)scaleLen;
		Dimension d = viewPanel.getSize();
		
		double lastVal = (end-len-1 >= 0)? data.getVal(end-len-1) : 0.0, val = 0.0;
		
		for(int i = 0; i < len; i++){
			for(int j = 0; j < 1440; j++){
				if(end+i-len >= 0){
					val = data.getVal(end+i-len, j);
					g.drawLine((int)((i*1440.0+j)*d.width/(len*1440.0)), (int)((1.0-lastVal)*(d.height-23))+2, (int)((i*1440.0+j+1)*d.width/(len*1440.0)), (int)((1.0-val)*(d.height-23))+2);
					lastVal = val;
				}
			}
		}
	}

	/** Plots a daily data set */
	protected void plotDays(GraphData data, Graphics g){
		
		int end = (int)(sim.minute()/1440.0-scaleEnd);
		int len = (scaleLen <= 0.0)? (int)(sim.minute()/1440.0-scaleEnd) : (int)(scaleLen);
		Dimension d = viewPanel.getSize();
		
		double lastVal = (end-len-1 >= 0)? data.getVal(end-len-1) : 0.0, val = 0.0;
		
//System.out.println("end = "+end+" len="+len+" plotting "+(end-len)+" to "+end);
		
		for(int i = 0; i < len; i++){
			if(end+i-len >= 0){
				val = data.getVal(end+i-len);
//System.out.println(""+(end+i-len)+":"+val);
//System.out.println("plotting point"+((i-1)*d.width/len)+","+((int)((1.0-lastVal)*d.height)+1)+" lastVal="+lastVal+" val="+val);
				g.drawLine(i*d.width/len, (int)((1.0-lastVal)*(d.height-23))+2, (i+1)*d.width/len, (int)((1.0-val)*(d.height-23))+2);
				lastVal = val;
			}
		}
		
//		for(int i = 0; i < 50; i++){
//			val = data.getVal(i);
//			System.out.println(""+i+":"+val);
//		}

	}

	
	/** Draws the axis */
	private void drawAxis(Graphics g){
		
		int end = (int)(sim.minute()/1440.0-scaleEnd+0.5);
		int len = (scaleLen <= 0.0)? (int)(sim.minute()/1440.0+0.5) : (int)(scaleLen+0.5);
		Dimension d = viewPanel.getSize();
			
		g.setColor(Color.BLACK);
		g.drawLine(0, d.height-20, d.width, d.height-20);
		
		int interval = (len < 10)? 1 : len/10;
		
			
		for(int i = 1; i < len; i++){
			if(i % interval == 0){
				g.drawLine(i*d.width/len, d.height-20, i*d.width/len, d.height-15);
				int val = end+i-len;
				int digits = (val >= 0)? (int)(Math.log(val)/Math.log(10)+1.0) : (int)(Math.log(-val)/Math.log(10)+2.0);
				g.drawString(""+val, i*d.width/len-4*digits, d.height-3);
			}else{
				g.drawLine(i*d.width/len, d.height-20, i*d.width/len, d.height-18);
			}
		}
	}
	
}
