/*
 * File:   Simulation.java
 * Author: Conrad Rider
 * Email:  c.rider@eviltree.co.uk
 * Date:   22-02-05
 */


package sim;

import gui.*;
import gui.view.*;
import utils.*;
import main.*;

import sim.pasture.*;
import sim.livestock.*;
import main.PastureSim;

import java.io.*;
import java.util.*;
import javax.swing.JOptionPane;
import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import javax.imageio.ImageIO;


import javax.swing.SwingUtilities;


/** This is the top level simulation object. All parts of the simulation
  * hold a reference to and interact with this object. */
public class Simulation {

	// This is the maximum speed the simulation is allowed to run at
	// Most computers cant simulate at this speed, but this could
	// be increased if using a powerful computer.
	public static final int MAX_SPEED = 4096;
	
	// The graphic is no longer interesting past 32X Zoom
	public static final int MAX_ZOOM = 32;
	
	// The simulation has been tuned so that each iteration
	// represents a minute of pasture life.
	public static int TIME_FACTOR = 60;
	
	// Sizes of objects in meters
 	public static final double ACTUAL_GATE_WIDTH = 2.6;
	public static final double ACTUAL_SNAP_RADIUS = 2.6;
 	public static final double ACTUAL_ANIMAL_SIZE = 1;
	
	// The modes the simulation can be in
	public static final int MODE_ADJUST               = 1;
	public static final int MODE_SIMULATE             = 0;
	public static final int MODE_PASTURE              = 100;
	public static final int MODE_PASTURE_ADD_FIELD    = 101;
	public static final int MODE_PASTURE_REMOVE_FIELD = 102;
	public static final int MODE_PASTURE_ADD_GATE     = 103;
	public static final int MODE_PASTURE_REMOVE_GATE  = 104;
	public static final int MODE_PASTURE_ADD_WATER    = 105;
	public static final int MODE_PASTURE_REMOVE_WATER = 106;
	public static final int MODE_LIVESTOCK            = 200;
	public static final int MODE_LIVESTOCK_ADD        = 201;
	public static final int MODE_LIVESTOCK_PLACE      = 202;
	public static final int MODE_LIVESTOCK_REMOVE     = 203;
	public static final int MODE_HISTORY              = 300;
	public static final int MODE_GRAPH                = 400;

	// The history state transition options
	public static final int STATE_START    = 0;
	public static final int STATE_JBACK    = 1;
	public static final int STATE_BACK     = 2;
	public static final int STATE_FORWARD  = 3;
	public static final int STATE_JFORWARD = 4;
	public static final int STATE_CURRENT  = 5;

	// The spacial data currently being viewed
	public static final int DATA_NONE       =  0;
	public static final int DATA_NITROGEN   = 10;
	public static final int DATA_CAPACITY   = 20;
	public static final int DATA_MOISTURE   = 30;
	public static final int DATA_HEIGHT     = 40;
	public static final int DATA_LEAF       = 50;
	public static final int DATA_FREQUENCY  = 60;
	public static final int DATA_WATER_DIST = 70;
	
	// Animal data currently being viewed
	public static final int DATA_NORMAL = 10;
	public static final int DATA_THIRST = 20;
	public static final int DATA_HUNGER = 30;
	
	private transient File simulationFile;       // The file holding the simulation
	private transient FileManager file;          // A reference to the file manager
	private transient BufferedImage arielPhoto;  // The aerial photograph representing the pasture
	private transient ActionPanel ap;            // The main view window of the user interface
	public  transient UserInterface ui;          // The whole user interface
	
	public transient State state; // The current state to be drawn (leve or old)
	private transient SimulationState curState; // The current live state
	
	// The number of pixels which represent a meter
	public transient double METER;
	public transient double METER_SQUARED;
	
	// The number of pixels which represent objects
 	public transient double GATE_WIDTH;
	public transient double SNAP_RADIUS;
 	public transient double ANIMAL_SIZE;
	
	public  transient int MAX_DIST; // The furthest distance in this simulation
	
	public  transient int mode;               // The current mode that it is running in
	private           boolean drawAriel;      // Whether to draw the arial shot
	private           boolean drawFields;     // Whether to draw the field boundaries, gates and water
	private           boolean drawDeadstock;  // Whether to draw the dead animals in the simulation
	private           boolean drawWaterPaths; // Whether to draw the paths to the water
	private           int drawSpacialData ;   // Which set of spacial data to draw, if any
	private           int drawAnimalData;     // Which set of animal data to draw, if any
	private transient long speed;             // The current speed its running at
	public  transient boolean maxSpeed;       // Wheather we are running at max speed
	
	private File arielPhotoFile;    // The file holding the ariel photo
	
	private int area;               // The user defined area of the pasture
	private int zoom;               // The current level of zoom
	private int panX;               // The current level of pan in the x axis
	private int panY;               // the current level of pan in the y axis	
	
	private long rate;              // The rate in saves per hour that the simulation is saved at
	private int history;            // The current history item we are browsing
	private long lastSave;          // The number of milliseconds since the last save
	
	private Dimension size;         // The dimensions in pixels of the pasture area
	
	public GraphData[] animalGraphData;
	public GraphData[] pastureGraphData;

	/** Constructor for a new simulation */
	public Simulation(File simulationFile, File arielPhotoFile, int area, long rate, UserInterface ui) throws Exception{
		this.simulationFile = simulationFile;
		

		// Create data dir
		String dir = simulationFile.toString();
		dir = dir.substring(0,dir.length()-4)+"_data";
		new File(dir).mkdir();//if(!new File(dir).mkdir()) throw new IOException("Failed to create data directory");
		
		this.arielPhotoFile = arielPhotoFile;
		this.ui = ui;
		//this.arielPhoto = ImageIO.read(arielPhotoFile);
		this.arielPhoto = ui.loadImage(arielPhotoFile, 0);
		this.size = new Dimension(arielPhoto.getWidth(null), arielPhoto.getHeight(null));
		float resolution = (size.width * size.height)/(float)area; 
		if(resolution < 1.0f){
			throw new IOException("Resolution is below 1 pixel per m\u00B2\n"+
					"With the given area, the simulation has\na "+
					"resolution of "+resolution+" pixels per m\u00B2");
		}
		this.area = area;
		this.rate = rate;
		this.zoom = InitConfig.zoom;
		this.panX = 0;
		this.panY = 0;
		this.lastSave = 0;
		
		this.drawAriel       = InitConfig.drawAriel;
		this.drawFields      = InitConfig.drawFields;
		this.drawDeadstock   = InitConfig.drawDeadstock;
		this.drawWaterPaths  = InitConfig.drawWaterPaths;
		this.drawSpacialData = DATA_NONE;
		this.drawAnimalData  = DATA_NORMAL;
		
		this.state = new SimulationState(this, arielPhotoFile);   // Create the first state
		this.curState = (SimulationState)state;
		this.file = new FileManager(simulationFile, this);        // Create this simulation's file manager
		file.createSimulation(((SimulationState)state).freeze()); // Writes the initial data to disk
		
		initialise();     // Initialises the simulation data and constants
		startSimThread(); // Starts the simulation thread
	}
	
	
	/** Constructor for an opened simulation */
	public Simulation(File simulationFile, UserInterface ui) throws Exception{
		this.simulationFile = simulationFile;
		this.ui = ui;
		this.file = new FileManager(simulationFile, this); // Create the file manager
		
		state = file.loadSimulation();                     // Use it to load the simulation from the file name

		initialise(); // Initialise sim values
				
		//((FrozenState)state).initialise(this);   // Initialise this frozen state
		state = ((FrozenState)state).thaw(this); // Make the state live
		curState = (SimulationState)state;

		//((SimulationState)state).setSimulation(this);
		startSimThread(); // Begin the simulation thread

		if(main.PastureSim.textOnly){ speed = main.PastureSim.speed; run(); }
	}

	
	/** Initialise default simulation settings on create or load */
	public void initialise(){
		//if(state instanceof SimulationState) this.curState = (SimulationState)state;
		this.mode = InitConfig.mode;
		this.speed = InitConfig.speed;
		this.history = file.historyCount(); //curState = (SimulationState)state;
		ui.setTitle(simulationFile.getName()+" - PastureSim");
		// Get the number of pixels per meter
		METER_SQUARED = (double)(size.width*size.height)/(double)area;
		METER = Math.sqrt(METER_SQUARED);
		
		GATE_WIDTH = ACTUAL_GATE_WIDTH*METER;
		SNAP_RADIUS = ACTUAL_SNAP_RADIUS*METER;
		ANIMAL_SIZE = ACTUAL_ANIMAL_SIZE*METER;
		
		MAX_DIST = (size.width<<1) + (size.height) >> 1;
	}

		
	/** Starts the simulation thread */
	public void startSimThread(){
		//state.setPriority(Thread.MAX_PRIORITY);
		state.start(); // Start the simulation thread
	}

		
	public boolean saving(){ return file.isSaving(); }
/*	public boolean enoughMemory(){
		// photograph consumpstion is proportional
		// to area * 4. There are two photographs
		// and an animal array of the same size
		Dimension d;
		if(arielPhoto != null) d = getPastureSize();
		else d = new Dimension(100, 100);
		long est = d.width * d.height * 10 + 100000;
System.out.println("FREE MEM: "+Runtime.getRuntime().freeMemory()+" (NEED " + est+")");
		long consumed = Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
//		System.out.println("CONSUMED = " + consumed + "(est "+est+")  FREE = "+Runtime.getRuntime().freeMemory());
		return Runtime.getRuntime().freeMemory() > est;
	
	}
*/	
/*	public boolean historyReady(){
		return file.doneLoad();
	}
*/
	public File nextSoilImageFile(){
		String imagePath = simulationFile.getParent();
		String imageName = simulationFile.getName();
		imageName = imageName.substring(0,imageName.length()-4);
		return new File(imagePath+File.separator+imageName+"_data"+File.separator+"soil_data"+file.historyCount()+".png");
	}
	public File nextPlantImageFile(){
		String imagePath = simulationFile.getParent();
		String imageName = simulationFile.getName();
		imageName = imageName.substring(0,imageName.length()-4);
		return new File(imagePath+File.separator+imageName+"_data"+File.separator+"plant_data"+file.historyCount()+".png");
	}
	
	public BufferedImage loadImage(File imgFile, int id){
		try{
			return ui.loadImage(imgFile, id);
		}catch(Exception e){
			e.printStackTrace();
		}
		return null;
	}
	
/*	private void storeArielPhoto(Image image){
		GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
		GraphicsDevice gs = ge.getDefaultScreenDevice();
		GraphicsConfiguration gc = gs.getDefaultConfiguration();
    
 		
   		arielPhoto = gc.createCompatibleImage(image.getWidth(null), image.getHeight(null), Transparency.BITMASK);
   		//arielPhoto = gc.createCompatibleImage(image.getWidth(null), image.getHeight(null), Transparency.OPAQUE);
		
		Graphics g = arielPhoto.createGraphics();
		g.drawImage(image, 0, 0, null);
	}
*/	
	// Set pointers to essential gui components
	public void setActionPanel(ActionPanel ap){ this.ap = ap; }
	
	// Refresh gui components
	public void refreshViewPanel(){ ap.repaint(); }
	public void refreshContextPanel(){ ui.refreshContextItems(); }
	public void refreshMenuPanel(){ ui.refreshMenu(); }

	// Field accessor methods	
	public int getHistoryNumber(){ return history; }
	public int getHistoryCount(){ return file.historyCount(); }
	public File getSimulationFile(){ return simulationFile; }
	public int getArea(){ return area; }
	public double getRainfall(){ return state.getRainfall(); }
	public double getTemperature(){ return state.getTemperature(); }
	public double getRadiation(){ return state.getRadiation(); }
	public double getTranspiration(){ return state.getTranspiration(); }
	public void setRainfall(double r){ state.setRainfall(r); }
	public void setTemperature(double t){ state.setTemperature(t); }
	public void setRadiation(double r){ state.setRadiation(r); }
	public void setTranspiration(double t){ state.setTranspiration(t); }
	
	
	/** Returns the day's time in minutes */
	public int dayMinute(){ return (int)(state.getTime() / 60000) % 1440; }
	
	/** Returns the total time in minutes */
	public long minute(){ return (int)(state.getTime() / 60000); }

	
		
	public long getRate(){ return rate; }
	public long getSpeed(){ return speed; }
	public boolean canSpeedUp(){
		if(state instanceof SimulationState){
			return speed < MAX_SPEED &&
			speed <= ((SimulationState)state).getLoopSpeed() && !maxSpeed;
		}
		return false;
	}
	public boolean canSlowDown(){ return speed > 0 || maxSpeed; }
	public void plusSpeed(){
		if(canSpeedUp()){
			if(speed == 0){ speed = 1; run(); }
			else speed *= 2;
		}
		refreshContextPanel();
	}
	public void maxSpeed(){
		if(!maxSpeed){
			maxSpeed = true;
			if(speed == 0){ run(); }
			speed = 1;
			refreshContextPanel();
		}
	}
	
	public void minusSpeed(){
		if(canSlowDown()){
			if(maxSpeed){ maxSpeed = false; speed = 0; stop(); }
			else if(speed == 1){ speed = 0; stop(); }
			else speed /= 2;
			refreshContextPanel();
		}
	}
	
	public Animal[][] getPos(){ return state.getPos(); }
	public int getZoom(){ return zoom * 100; }
	public boolean canZoomIn(){ return zoom < MAX_ZOOM; }
	public boolean canZoomOut(){ return zoom > 1; }
	public void plusZoom(){
		if(zoom < MAX_ZOOM){
			zoom *= 2;
		 	//GATE_WIDTH *= 2; 
			//GATEPOST_DIAMETER *= 2;
			//SNAP_RADIUS *= 2;			
			panX *= 2;
			panY *= 2;
			pan(0, 0);
			ui.refreshMenuItems();
			ui.refreshContextItems();
			ap.repaint();
		}
	}
	public void minusZoom(){
		if(zoom > 1){
			zoom /= 2;
		 	//GATE_WIDTH /= 2; 
			//GATEPOST_DIAMETER /= 2;
			//SNAP_RADIUS /= 2;			
			panX /= 2;
			panY /= 2;
			pan(0, 0);
			ui.refreshMenuItems();
			ui.refreshContextItems();
			ap.repaint();
		}
	}
	public void pan(int dX, int dY){
		Dimension s = getScaledSize();
		Dimension d = getPastureSize();
		     if(panX + dX < -s.width/2 + d.width/2) panX = -s.width/2 + d.width/2;
		else if(panX + dX >  s.width/2 - d.width/2) panX =  s.width/2 - d.width/2;
		else panX += dX; 
		
		     if(panY + dY < -s.height/2 + d.height/2) panY = -s.height/2 + d.height/2;
		else if(panY + dY >  s.height/2 - d.height/2) panY =  s.height/2 - d.height/2;
		else panY += dY; 
	}
	public void drawScaledLine(Graphics g, double x1, double y1, double x2, double y2){
		Point p1 = scale(x1, y1); Point p2 = scale(x2, y2);
		g.drawLine(p1.x, p1.y, p2.x, p2.y);
	}
	public void fillScaledOval(Graphics g, double x, double y, double w, double h){
		Point p = scale(x, y);
		g.fillOval(p.x, p.y, (int)(w*zoom) , (int)(h*zoom));
	}
	public void fillScaledRect(Graphics g, double x, double y, double w, double h){
		Point p = scale(x, y);
		g.fillRect(p.x, p.y, (int)(w*zoom) , (int)(h*zoom));
	}
	public void fillScaledPolygon(Graphics g, int[] xPoints, int[] yPoints, int nPoints){
		Point p;
		int[] xp = new int[nPoints];
		int[] yp = new int[nPoints];
		for(int i = 0; i < nPoints; i++){
			p = scale(xPoints[i], yPoints[i]);
			xp[i] = Math.round(p.x);
			yp[i] = Math.round(p.y);
		}
		g.fillPolygon(xp, yp, nPoints);
	}	
	public Point scale(double x, double y){ return scale(new Point2D.Double(x,y)); }
	public Point scale(Point2D.Double p){
		Dimension s = getScaledSize();
		Dimension d = getPastureSize();
		return new Point(
			(int)(((p.x-d.width/2)*(zoom)) + d.width/2 + panX + 0.5),
			(int)(((p.y-d.height/2)*(zoom)) + d.height/2 + panY+0.5));
	}
	
	public Point getTopLeft(){
		Dimension s = getScaledSize();
		Dimension d = getPastureSize();
		return new Point(panX-((s.width/2)-(d.width/2)), panY-((s.height/2)-(d.height/2)));
	}
	
	public Point2D.Double scaledMousePos(int xi, int yi){
		double x = (double)xi-0.5;
		double y = (double)yi-0.5;
		Dimension d = getPastureSize();
		return new Point2D.Double((x-d.width/2.0-panX)/zoom + d.width/2.0, (y-d.height/2.0-panY)/zoom + d.height/2.0);
	}
	
	public int getMode(){ return mode; }
	
	
	// Pasture specific methods
	
	public Dimension getPastureSize(){
		return size;
	}
	public Dimension getScaledSize(){
		return new Dimension(
			size.width * zoom,
			size.height * zoom);
	}
	// Returns true if the coordinates are within the field bounds, false otherwise
	public boolean withinPastureBounds(int x, int y){
		return x >= 0 && y >= 0 && x < size.width && y < size.height;
	}
	
	
	public void reDraw(PastureState p){
		if(drawSpacialData  == DATA_NITROGEN)  p.filterN();
		if(drawSpacialData  == DATA_CAPACITY)  p.filterC();
		if(drawSpacialData  == DATA_MOISTURE)  p.filterW();
		if(drawSpacialData  == DATA_HEIGHT)    p.filterH();
		if(drawSpacialData  == DATA_LEAF)      p.filterA();
		if(drawSpacialData  == DATA_FREQUENCY) p.filterV();
		refreshContextPanel();
		// Refresh graphs every day
		if(minute() % 1440 == 0) ui.refreshGraphs();
		ap.repaint();
		if(!PastureSim.printPastureData && !PastureSim.printAnimalData) System.out.print(".");
	}
	
	public void draw(){ ap.repaint(); } //ui.renderOnce(); }//
	public void drawPasture(Graphics g){
		//System.out.print("Drawing...");
		Dimension s = getScaledSize();
		Point p = getTopLeft();
		if(drawSpacialData  == DATA_NONE){
			if(drawAriel) g.drawImage(arielPhoto, p.x, p.y, s.width, s.height, null);
		}else{
			state.drawData(g, p.x, p.y, s.width, s.height, drawSpacialData );
			/*
			No need to super impose the aerial photo
			if(drawAriel && (drawSpacialData  == DATA_FREQUENCY)){
				ImageFilter filter = new AlphaFilter();
				Image img = Toolkit.getDefaultToolkit().createImage(new FilteredImageSource(arielPhoto.getSource(), filter));
				g.drawImage(img, p.x, p.y, s.width, s.height, null);
			}
			*/
		}		

		if(drawFields){ state.drawFields(g); }
		if(drawWaterPaths){ state.drawWaterPaths(g); }
		if(drawDeadstock){ state.drawDeadstock(g); }
		if(drawAnimalData != DATA_NONE){ state.drawLivestock(g); }
		//if(mode == MODE_SIMULATE) ui.refreshContextItems();
	}
	
	public void drawActiveAnimal(Graphics g, Point2D.Double p){
		g.setColor(Animal.COLOR_ACTIVE);
		fillScaledOval(g, p.x-ANIMAL_SIZE/2, p.y-ANIMAL_SIZE/2, ANIMAL_SIZE, ANIMAL_SIZE);	
	}
	
	
	public void addField(FieldBounds f){ state.add(f); }
	public void removeField(FieldBounds f){ state.remove(f); }
	public boolean hasFields(){ return state.hasFields(); }
	public FieldGeometry getFieldGeometry(){ return state.getFieldGeometry(); }
	
	public void addGate(FieldGate g){ state.add(g);	}
	public void removeGate(FieldGate g){ state.remove(g); }
	public boolean hasGates(){ return state.hasGates(); }
	public GateGeometry getGateGeometry(){ return state.getGateGeometry(); }
	public void notifyGateSwitch(){ state.notifyGateSwitch(); }
	
	public void addWater(FieldWater w){ state.add(w);	}
	public void removeWater(FieldWater w){ state.remove(w); }
	public boolean hasWater(){ return state.hasWater(); }
	public WaterGeometry getWaterGeometry(){ return state.getWaterGeometry(); }
	public boolean toDrawWaterPaths(){ return drawWaterPaths; }
	public boolean toDrawWaterDist(){ return drawSpacialData  == DATA_WATER_DIST; }
	public PathFinderMap getWaterPath(){ return state.getWaterPath(); }
	
	// Livestock specific methods
	
	public void addLivestock(LivestockGroup g){ state.add(g); }
	public LinkedList getLivestock(){ return state.getLivestock(); }
	public void removeLivestock(LivestockGroup g){ state.remove(g); }
	public boolean hasLivestock(){ return state.hasLivestock(); }
	
		
	// Menu selection handlers
	
	public void saveAs(){
		File newFile = FileSelector.saveSimulation(ui);
		if(newFile != null){
			FileManager newManager = file.copySimulation(newFile);
			if(newManager != null) file = newManager;
			simulationFile = newFile;
			
			// Create data dir
			String dir = simulationFile.toString();
			dir = dir.substring(0,dir.length()-4)+"_data";
			new File(dir).mkdir();//if(!new File(dir).mkdir()) throw new IOException("Failed to create data directory");
			
			// Set title
			ui.setTitle(simulationFile.getName()+" - PastureSim");
			
			// Save it
			save();
		}
	}
	
	public void save(){ 
	
		FrozenState f = ((SimulationState)curState).freeze();
		//f.loadPastureStateImage(this);
		file.saveSimulation(f);

	}
	
	public void export(){
		File f = FileSelector.saveImage(ui);
		if(f != null) file.export(f);
	}
	
	public void close(){
		//save();
		file.close();
		((SimulationState)curState).kill();
		ui.dispose();
		PastureSim.cleanUp(ui);
	}		

	public void setMode(int mode){
		if(this.mode != mode){
			this.mode = mode;
			if(mode != MODE_SIMULATE) stop();
			if(mode != MODE_HISTORY){
				state = curState;
				history = file.historyCount();
			}
			ui.switchContext();
			if(mode == MODE_SIMULATE && speed > 0) run();
		}
	}
	
	
	public void saveHistory(long totalTime){
		long interval = rate*3600000l;
		if(rate > 0l && (lastSave + interval) < totalTime){
			lastSave = totalTime;
			file.saveSimulation((FrozenState)((SimulationState)curState).freeze());
		}
	}
	public boolean canViewHistory(int pos){
		switch(pos){
			case STATE_START: return history > 0;
			case STATE_JBACK: return history >= 10;
			case STATE_BACK: return history > 0;
			case STATE_FORWARD: return history < file.historyCount();
			case STATE_JFORWARD: return history < file.historyCount()-10;
			case STATE_CURRENT: return history < file.historyCount();
		}
		return false;
	}
	public void gotoState(int pos){
//System.out.println("Can goto pos:"+pos+" = "+canViewHistory(pos));
		
		// We can only go to different states if there are others
//System.out.print("history size = " + previousStates.size());
		int oldHist = history;	
		// Update the history position counter
		switch(pos){
			case STATE_START:
				if(canViewHistory(STATE_START)) history = 0;
				break;
			case STATE_JBACK:
				if(canViewHistory(STATE_JBACK)) history -= 10;
				else if(canViewHistory(STATE_BACK)) history = 0;
				break;
			case STATE_BACK:
				if(canViewHistory(STATE_BACK)){
//System.out.println("history = " + history);
					history--;
//System.out.println("history decramendtd");
//System.out.println("history = " + history);
				}
				break;
			case STATE_FORWARD:
				if(canViewHistory(STATE_FORWARD)) history++;
				break;
			case STATE_JFORWARD:
				if(canViewHistory(STATE_JFORWARD)) history += 10;
				else if(canViewHistory(STATE_FORWARD)) history = file.historyCount();
				break;
			case STATE_CURRENT:
				if(canViewHistory(STATE_CURRENT)) history = file.historyCount();
				break;
		}
	
		// Show the state
//System.out.println("Viewing history: "+history);
		if(history == file.historyCount()){
			state = curState;
		}
		else{
			State s = file.loadHistory(history);
			if(s != null){
				state = s;
				((FrozenState)state).initialise(this);
			}else{
				history = oldHist;
			}
		}
	}
	
	public void toggleViewAriel()     { drawAriel      = !drawAriel;      ap.repaint(); }
	public void toggleViewFields()    { drawFields     = !drawFields;     ap.repaint(); }
	public void toggleViewDeadstock() { drawDeadstock  = !drawDeadstock;  ap.repaint(); }
	public void toggleViewWaterPaths(){ drawWaterPaths = !drawWaterPaths; ap.repaint(); }
	public boolean viewingAriel()     { return drawAriel;      }
	public boolean viewingFields()    { return drawFields;     }
	public boolean viewingDeadstock() { return drawDeadstock;  }
	public boolean viewingWaterPaths(){ return drawWaterPaths; }
	public void showData(int type){ drawSpacialData  = type; ap.repaint(); }
	public boolean showingNitrogen() { return drawSpacialData  == DATA_NITROGEN; }
	public boolean showingCapacity() { return drawSpacialData  == DATA_CAPACITY; }
	public boolean showingMoisture() { return drawSpacialData  == DATA_MOISTURE; }
	public boolean showingHeight()   { return drawSpacialData  == DATA_HEIGHT; }
	public boolean showingLeaf()     { return drawSpacialData  == DATA_LEAF; }
	public boolean showingFrequency(){ return drawSpacialData  == DATA_FREQUENCY; }
	public boolean showingProximity(){ return drawSpacialData  == DATA_WATER_DIST; }
	public boolean showingNone()     { return drawSpacialData  == DATA_NONE; }
	
	public void showAnimalData(int type){ drawAnimalData  = type; ap.repaint(); }
	public boolean showingAnimalThirst(){ return drawAnimalData  == DATA_THIRST; }
	public boolean showingAnimalHunger(){ return drawAnimalData  == DATA_HUNGER; }
	public boolean showingAnimalNormal(){ return drawAnimalData  == DATA_NORMAL; }
	public boolean showingNoAnimals()   { return drawAnimalData  == DATA_NONE; }
	
	public PastureState getPastureState(){ return state.getPastureState(); }
	
	
	public void loadSimulation(ObjectInputStream objIn) throws Exception{

		this.arielPhotoFile = (File)objIn.readObject();
		this.arielPhoto = ui.loadImage(arielPhotoFile, 0);
		//this.arielPhoto = ImageIO.read(arielPhotoFile);
		//this.curState = (SimulationState)objIn.readObject();
		//this.curState = new SimulationState(objIn, this);
		this.area = objIn.readInt();
		this.zoom = objIn.readInt();
		this.panX = objIn.readInt();
		this.panY = objIn.readInt();
	
		this.rate     = objIn.readLong();
		this.history  = file.historyCount();
		this.lastSave = objIn.readLong();
	
		this.drawAriel       = objIn.readBoolean();
		this.drawFields      = objIn.readBoolean();
		this.drawDeadstock   = objIn.readBoolean();
		this.drawWaterPaths  = objIn.readBoolean();
		this.drawSpacialData = objIn.readInt();
		this.drawAnimalData  = objIn.readInt();
		this.size            = (Dimension)objIn.readObject();
		
		this.animalGraphData = (GraphData[])objIn.readObject();
		this.pastureGraphData = (GraphData[])objIn.readObject();
		
		/*
		System.out.println("Read object "+arielPhotoFile+" - [arielPhotoFile]");
		System.out.println("Read object "+previousStates+" - [previousStates]");
		System.out.println("Read object "+state+" - [state]");
		System.out.println("Read object "+area+" - [area]");
		System.out.println("Read object "+rate+" - [rate]");
		System.out.println("UI = " + ui);
		*/
		/*
		storeArielPhoto(ui.loadImage(arielPhotoFile.toString()));
		if(arielPhoto == null || arielPhoto.getWidth(null)<=0){
			throw new IOException("Could not read image file.\n\nFILE: " + arielPhotoFile);
		}
		*/
	}
	
	public void saveSimulation(ObjectOutputStream objOut) throws Exception{
		
		objOut.writeObject(arielPhotoFile);
//System.out.println("ATTEMPTIMG TO WRITE STATE OBJ:" + curState);
//curState.writeData(objOut);
		//objOut.writeObject(curState);
		
		objOut.writeInt(area);
		objOut.writeInt(zoom);
		objOut.writeInt(panX);
		objOut.writeInt(panY);
	
		objOut.writeLong(rate);
		objOut.writeLong(lastSave);
	
		objOut.writeBoolean(drawAriel);
		objOut.writeBoolean(drawFields);
		objOut.writeBoolean(drawDeadstock);
		objOut.writeBoolean(drawWaterPaths);
		objOut.writeInt(drawSpacialData);
		objOut.writeInt(drawAnimalData);
		objOut.writeObject(size);
		
		objOut.writeObject(animalGraphData);
		objOut.writeObject(pastureGraphData);
		/*
		System.out.println("Wrote object "+arielPhotoFile);
		System.out.println("Wrote object "+previousStates);
		System.out.println("Wrote object "+curState);
		System.out.println("Wrote object "+area);
		System.out.println("Wrote object "+rate);
		*/
	}
	
	public void run(){  
		if(state instanceof SimulationState)
			((SimulationState)state).play();
	}//ui.startRender(); }

	public void stop(){
		if(state instanceof SimulationState){
			((SimulationState)state).pause();
			ui.refreshMenu(); 
		}
	}//ui.stopRender(); }
		
	//private void writeObject(java.io.ObjectOutputStream out) throws IOException{}
	//private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException{}

	
	
	class AlphaFilter extends RGBImageFilter{
		
		private static final int ALPHA_LEVEL = 0x50000000;
		
		private static final int RED_MASK = 0x00FF0000;
		private static final int GREEN_MASK = 0x0000FF00;
		private static final int BLUE_MASK = 0x000000FF;
		
		public int filterRGB(int x, int y, int rgb){
			int red = (rgb & RED_MASK) >> 16;
			int green = (rgb & GREEN_MASK) >> 8;
			int blue = rgb & BLUE_MASK;
	
			int pixel = (int)((red+green+blue)/3);
			return ALPHA_LEVEL | (pixel << 16) | (pixel << 8) | pixel;
		}
	}
	
}
