/*
 * File:   State.java
 * Author: Conrad Rider
 * Email:  c.rider@eviltree.co.uk
 * Date:   22-02-05
 */


package sim;

import sim.pasture.*;
import sim.livestock.*;
import gui.*;
import utils.*;

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.io.*;
import java.util.*;
import javax.swing.JOptionPane;


/** This is the super class of a live or frozen simulation state */
public abstract class State extends Thread implements AnimalContainer{

	protected transient Simulation sim;          // A reference to the simulation
	protected transient PathFinderMap waterPath; // The path finder map
	protected transient LinkedList ownedGates;   // The gates which we own
	
	protected LinkedList fields;   // The fields
	protected LinkedList gates;    // The gates
	protected LinkedList water;    // The water
	protected LinkedList animals;  // The animals
	protected PastureState pState; // The the spacial pasture state data
	protected Animal[][] pos;      // The array of animal positions
	protected long totalTime;      // The total amount of time which has passed in milliseconds
	
	/** Initialises all state objects after a load */
	public void initialise(Simulation sim){
		this.sim = sim;
		pState.initialise(sim);
		pos = new Animal[sim.getPastureSize().width][sim.getPastureSize().height];
		
		for(Iterator i = fields.iterator(); i.hasNext();){
			((FieldBounds)i.next()).initialise(sim, gates);
		}
		for(Iterator i = gates.iterator(); i.hasNext();){
			((FieldGate)i.next()).initialise(sim);
		}
		for(Iterator i = water.iterator(); i.hasNext();){
			((FieldWater)i.next()).initialise(sim);
		}
		for(Iterator i = animals.iterator(); i.hasNext();){
			((LivestockGroup)i.next()).initialise(sim, pos, pState);
		}
		
		ownedGates = new LinkedList();
		for(Iterator i = gates.iterator(); i.hasNext();){
			FieldGate g = (FieldGate)i.next();
			if(g.neighbour(this) != null) ownedGates.add(g);
		}
	
		waterPath = new PathFinderMap(sim);
	}
	
	public LinkedList getLivestock(){ return animals; }
	public LinkedList getFields(){ return fields; }
	public Animal[][] getPos(){ return pos; }
	public PastureState getPastureState(){ return pState; }

	
	public long getTime(){ return totalTime; }
	public String getReadableTime(){
		
		String t = "";
		if(totalTime >= 315792000000l){   // A year in milliseconds
			t += ((int)(totalTime / 315792000000l))+"Y, ";
			t += ((int)((totalTime  / 86400000) % 365.25))+"D, ";
		}
		else if(totalTime >= 2629800000l){ // A month in milliseconds
			t += ((int)(totalTime / 2629800000l))+"M, ";
			t += ((int)((totalTime  / 86400000) % 30.4375))+"D, ";
		}else if(totalTime >= 604800000){  // A week in milliseconds
			t += ((int)(totalTime / 604800000))+"W, ";
			t += ((int)((totalTime  / 86400000) % 7))+"D, ";
		}else if(totalTime >= 86400000){// A day in milliseconds
			t += ((int)(totalTime  / 86400000))+"D, ";
		}
		int hour = (int)(totalTime  / 3600000) % 24;
		if(hour < 10) t += " 0"+hour;
		else          t +=      hour;
		int min =  (int)(totalTime  / 60000) % 60;
		if(min < 10)  t += ":0"+min;
		else          t += ":" +min;
		int sec =  (int)(totalTime  / 1000) % 60;
		if(sec < 10)  t += ":0" +sec;
		else          t += ":"  +sec;
		//int mil = (int)(totalTime % 1000);
		//if(mil < 10)       t += ".00"+mil;
		//else if(mil < 100) t += ".0"+mil;
		//else               t += "."+mil;
		return t;
	}


	// Methods for drawing state
	public void drawFields(Graphics g){
		for(Iterator i = water.iterator(); i.hasNext();){
			((FieldWater)i.next()).draw(g);
		}
		for(Iterator i = fields.iterator(); i.hasNext();){
			((FieldBounds)i.next()).draw(g);
		}
		for(Iterator i = gates.iterator(); i.hasNext();){
			((FieldGate)i.next()).draw(g);
		}

	}
	
	// Method to draw water paths
	public void drawWaterPaths(Graphics g){
		if(waterPath != null){ waterPath.drawDirection(g); }
	}
	
	
	// Draw the livestock
	public void drawLivestock(Graphics g){
		if(sim.showingAnimalHunger()){
			for(Iterator i = animals.iterator(); i.hasNext();){
				((LivestockGroup)i.next()).drawHunger(g);
			}
		}else if(sim.showingAnimalThirst()){
			for(Iterator i = animals.iterator(); i.hasNext();){
				((LivestockGroup)i.next()).drawThirst(g);
			}
		}else{
			for(Iterator i = animals.iterator(); i.hasNext();){
				((LivestockGroup)i.next()).draw(g);
			}
		}
	}
	
	// Draw the deadstock
	public void drawDeadstock(Graphics g){
		for(Iterator i = animals.iterator(); i.hasNext();){
			((LivestockGroup)i.next()).drawDead(g);
		}
	}
		
	// Methods for displaying state data
	public void drawData(Graphics g, int x, int y, int width, int height, int dataType){
		switch(dataType){
			case Simulation.DATA_NITROGEN:
				pState.drawNitrogen(g, x, y, width, height);
				break;
			case Simulation.DATA_CAPACITY:
				pState.drawCapacity(g, x, y, width, height);
				break;
			case Simulation.DATA_MOISTURE:
				pState.drawWater(g, x, y, width, height);
				break;
			case Simulation.DATA_HEIGHT:
				pState.drawGrassHeight(g, x, y, width, height);
				break;
			case Simulation.DATA_LEAF:
				pState.drawLeafArea(g, x, y, width, height);
				break;
			case Simulation.DATA_FREQUENCY:
				pState.drawVisitFreq(g, x, y, width, height);
				break;
			case Simulation.DATA_WATER_DIST:
				waterPath.drawDistance(g);
		}
	}
	
	
	public void add(FieldBounds f){
		fields.add(f);
		pState.newBounds(f);
		f.addExistingGates(gates);
		for(Iterator i = animals.iterator(); i.hasNext();){
			((LivestockGroup)i.next()).notifyNewField(f);
		}
		waterPath = new PathFinderMap(sim);		
	}
	public void add(FieldGate g){
		for(Iterator i = fields.iterator(); i.hasNext();){
			((FieldBounds)i.next()).notifyGateArrival(g);
		}
		gates.add(g);
		if(g.neighbour(this) != null) ownedGates.add(g);
	}
	public void add(LivestockGroup g){ animals.add(g); }
	public void remove(FieldBounds f){
		fields.remove(f);
		pState.boundsRemoved(f);
		LinkedList gatesToRemove = new LinkedList();
		for(Iterator i = gates.iterator(); i.hasNext();){
			FieldGate g = (FieldGate)i.next();
			if(!g.hasField(fields)) gatesToRemove.add(g);
			else g.notifyFieldRemoval(f);
		}
		for(Iterator i = animals.iterator(); i.hasNext();){
			((LivestockGroup)i.next()).notifyFieldRemoval(f);
		}
		while(!gatesToRemove.isEmpty()){
			FieldGate g = (FieldGate)gatesToRemove.getFirst();
			gatesToRemove.remove(g); remove(g);
		}
		waterPath = new PathFinderMap(sim);		
	}
	public void remove(FieldGate g){
		for(Iterator i = fields.iterator(); i.hasNext();){
			((FieldBounds)i.next()).notifyGateRemoval(g);
		}
		gates.remove(g);
		ownedGates.remove(g);
		// If the gate was open we will need to update the water map
		if(g.isOpen()){ waterPath = new PathFinderMap(sim); }
	}
	// Tell the fields to update water maps when gates open or close
	public void notifyGateSwitch(){ waterPath = new PathFinderMap(sim); }
	
	public void add(FieldWater w){
		pState.newWater(w);
		water.add(w);
		waterPath = new PathFinderMap(sim);
	}
	public void remove(FieldWater w){
		pState.waterRemoved(w);
		water.remove(w);
		waterPath = new PathFinderMap(sim);
	}

	
	//-------------[ This implements AnimalContainer ]-------------

	public PathFinderMap getWaterPath(){ return waterPath; }
		
	// Returns wheather we own this area
	public boolean ownsArea(Point2D.Double p, int s){
		return !sim.getFieldGeometry().hasContainingField(p, s);
	}
	// Returns any connected fields containing this point over an open gate
	public AnimalContainer findOpenNeighbours(Point2D.Double p){
		for(Iterator i = ownedGates.iterator(); i.hasNext();){
			FieldGate g = (FieldGate)i.next();
			if(g.near(p)) return g.neighbour(this);
		}
		return null;
	}
	
	// This tells us wheather the animal container holds point p
	public boolean holds(Point2D.Double p){
//if(!sim.getFieldGeometry().hasContainingField(p)){System.out.println("we own the point");} 
	return !sim.getFieldGeometry().hasContainingField(p); }

	// This tells us wheather the animal container fully holds the pixel at p
	public boolean fullyHolds(Point2D.Double p){ return holds(p); }
	
	
	/** Returns true if the animal containers are connected by an open gate through the line l*/
	public boolean connected(AnimalContainer ac, Line2D.Double l){
		for(Iterator i = ownedGates.iterator(); i.hasNext();){
			FieldGate g = (FieldGate)i.next();
			if(g.isOpen() && g.neighbour(this) == ac && g.lineFits(l)) return true;
		}
		return false;
	}

	
	// Finds the gate leading to the shortest path to the target
	public FieldGate findGate(Point2D.Double tgt){
		AnimalContainer tgtBounds = sim.getFieldGeometry().findContainingField(tgt);
		if(tgtBounds == null) return null;
		
		TreeSet cand = tgtBounds.fringeEdges(tgt);
		LinkedList visited = new LinkedList();
		
		while(cand.size() > 0){
			DijkstraEdge first = (DijkstraEdge)cand.first(); // Get the closest candidate gate
			if(first.next == this) return first.gate; // If its connecting field is this one then we are done
			else{ 
				cand.remove(first); // If not, then remove it
				visited.add(first.gate); // add it to the visited list
				cand.addAll(first.next.fringeEdges(first, visited)); // Then add all its field's candidates
			}
		}
		return null;
	}
	
	
	// Returns all dijkstra edges (gates) which are reachable
	// from the point p through this field, ordered by distance
	public TreeSet fringeEdges(Point2D.Double p){
		TreeSet edges = new TreeSet();
		for(Iterator i = ownedGates.iterator(); i.hasNext();){
			FieldGate g = (FieldGate)i.next();
			if(g.isOpen()){
				edges.add(new DijkstraEdge(g, g.distance(p), this));
			}
		}
		return edges;
	}
	
	
	// Returns all dijkstra edges (gates) which are reachable
	// from the gate e through this field, ordered by distance
	public TreeSet fringeEdges(DijkstraEdge e, LinkedList visited){
		TreeSet edges = new TreeSet();
		for(Iterator i = ownedGates.iterator(); i.hasNext();){
			FieldGate g = (FieldGate)i.next();
			if(g.isOpen() && !visited.contains(g)){
				edges.add(new DijkstraEdge(g, g.distance(e.gate)+e.weight, this));
			}
		}
		return edges;
	}

	
		
	/** Returns weather the point p is on any of the field's gates */
	public boolean onGate(Point2D.Double p){
		for(Iterator i = ownedGates.iterator(); i.hasNext();){
			FieldGate g = (FieldGate)i.next();
			if(p.distance(g) < sim.GATE_WIDTH*0.5){
				return true;
			}
		}
		return false;
	}
	
	// Removes the livestock group
	public void remove(LivestockGroup g){ animals.remove(g); }
	
	public FieldGeometry getFieldGeometry(){ return new FieldGeometry(fields, sim); }
	public GateGeometry getGateGeometry(){ return new GateGeometry(fields, gates, sim); }
	public WaterGeometry getWaterGeometry(){ return new WaterGeometry(water, fields, sim); }
	
	public boolean hasFields(){ return fields.size() != 0; }
	public boolean hasGates(){ return gates.size() != 0; }
	public boolean hasWater(){ return water.size() != 0; }
	public boolean hasLivestock(){ return animals.size() != 0; }


	public double getRainfall(){ return pState.getRainfall(); }
	public double getTemperature(){ return pState.getTemperature(); }
	public double getRadiation(){ return pState.getRadiation(); }
	public double getTranspiration(){ return pState.getTranspiration(); }
	public void setRainfall(double r){ pState.setRainfall(r); }
	public void setTemperature(double t){ pState.setTemperature(t); }
	public void setRadiation(double r){ pState.setRadiation(r); }
	public void setTranspiration(double t){ pState.setTranspiration(t); }

}
