/*
 * File:   LivestockGroup.java
 * Author: Conrad Rider
 * Email:  c.rider@eviltree.co.uk
 * Date:   22-02-05
 */

 
 
package sim.livestock;

import sim.*;
import sim.livestock.*;
import sim.pasture.*;
import main.PastureSim;
import utils.GraphData;

import java.util.*;
import java.awt.*;
import java.io.*;
import java.awt.geom.*;


/** This represents groups or herds of livestock */
public class LivestockGroup extends LinkedList implements Serializable{

	private static final float MEM_DAMPEN_FACTOR = 1.1f;

	private transient Simulation sim;
	private transient int n;  // Number of animals
	private transient LinkedList toDie; // The animals which have been flaged to die
	private transient int listBeingModified; // Wheather a list is being modified
	private String name;
	private int type;
	private int state;	
	private LinkedList deceased; // The animals which have died
	public HerdMemory memory; // The herd's memory for good pastures
	
	/** Construct a new livestock group */
	public LivestockGroup(Simulation sim, String name, int type, int state, int n){
		this.sim = sim;
		this.name = name;
		this.type = type;
		this.state = state;
		this.n = n;
		this.deceased = new LinkedList();
		this.toDie = new LinkedList();
		this.memory = new HerdMemory();
	}
	
	/** Initialise global references after a load */
	public void initialise(Simulation sim, Animal[][] pos, PastureState pState){
		this.sim = sim;
		n = size();
		toDie = new LinkedList();
		for(Iterator i = iterator(); i.hasNext();){
			((Animal)i.next()).initialise(sim, pos, pState, this);
		}
		for(Iterator i = deceased.iterator(); i.hasNext();){
			((Animal)i.next()).initialise(sim, pos, pState, this);
		}
	}

	/** Initialise global references after a load */
	public void initialise(Simulation sim, PastureState pState){
		this.sim = sim;
		n = size();
		toDie = new LinkedList();
		for(Iterator i = iterator(); i.hasNext();){
			((Animal)i.next()).initialise(sim, pState, this);
		}
		for(Iterator i = deceased.iterator(); i.hasNext();){
			((Animal)i.next()).initialise(sim, pState, this);
		}
	}

	
	public void createAnimals(Point2D.Double pt){
		switch(type){
			case Animal.COW: for(int i = 0; i < n; i++){ add(new Cow(sim, pt.x,pt.y,this)); } break;
			case Animal.SHEEP: for(int i = 0; i < n; i++){ add(new Sheep(sim, pt.x,pt.y,this)); } break;
			case Animal.HORSE: for(int i = 0; i < n; i++){ add(new Horse(sim, pt.x,pt.y,this)); } break;
		}
	}
	
	public void notifyNewField(FieldBounds f){	
		for(Iterator i = iterator(); i.hasNext();){
			((Animal)i.next()).notifyNewField(f);
		}
	}
	public void notifyFieldRemoval(FieldBounds f){
		for(Iterator i = iterator(); i.hasNext();){
			((Animal)i.next()).notifyFieldRemoval(f);
		}
	}

	public void placeAnimals(Point2D.Double p){
		for(Iterator i = iterator(); i.hasNext();){
			((Animal)i.next()).place(p);
		}
	}
	
	public String getName(){ return name; }
	
	public int getSize(){ return n; }
	
	// Animates the animals and estimates the centre of the herd
	public void animate(){
		// Dampen memory of good grass
		if(sim.dayMinute() == 480) memory.dampen();
		
		int eat = 0;
		int drink = 0;
		int mem = 0;
		int lmem = 0;
		int herd = 0;
		int forage = 0;
		int nothing = 0;
		double energy = 0;
		double hydration = 0;
		
		// Animata animals
		for(Iterator i = iterator(); i.hasNext();){
			Animal a = (Animal)i.next();
			a.animate();
			if(a instanceof Sheep){
				switch(((Sheep)a).getDesicion()){
					case Animal.EAT       : eat++;     break;
					case Animal.DRINK     : drink++;   break;
					case Animal.MEMORY    : mem++;     break;
					case Animal.LMEMORY   : lmem++;    break;
					case Animal.HERD      : herd++;    break;
					case Animal.FORAGE    : forage++;  break;
					case Animal.DO_NOTHING: nothing++; break;
				}
				double en = (a.energy()-Sheep.goodFoodLevel)/Sheep.SHEEP_DAILY_FOOD;
				double hy = (a.hydration()-Sheep.goodWaterLevel)/Sheep.SHEEP_DAILY_WATER;
				energy += (en > 1.0)? 1.0 : (en < 0.0)? 0.0 : en;
				hydration += (hy > 1.0)? 1.0 : (hy < 0.0)? 0.0 : hy;
			}
		}
		
		if(sim.animalGraphData == null){
			sim.animalGraphData = new GraphData[9];
			for(int i = 0; i < 9; i++){ sim.animalGraphData[i] = new GraphData(); }
		}
		sim.animalGraphData[0].updateMin(sim.minute(), (double)eat/(double)n);
		sim.animalGraphData[1].updateMin(sim.minute(), (double)forage/(double)n);
		sim.animalGraphData[2].updateMin(sim.minute(), (double)drink/(double)n);
		sim.animalGraphData[3].updateMin(sim.minute(), (double)herd/(double)n);
		sim.animalGraphData[4].updateMin(sim.minute(), (double)mem/(double)n);
		sim.animalGraphData[5].updateMin(sim.minute(), (double)lmem/(double)n);
		sim.animalGraphData[6].updateMin(sim.minute(), (double)nothing/(double)n);
		double hunger = 1.0-((double)energy)/((double)n);
		double thirst = 1.0-((double)hydration)/((double)n);
		sim.animalGraphData[7].updateMin(sim.minute(), (hunger < 0.0)? 0.0 : hunger );
		sim.animalGraphData[8].updateMin(sim.minute(), (thirst < 0.0)? 0.0 : thirst );
		
		// Print stats
		if(PastureSim.printAnimalData){
			PastureSim.writeAnimal(str(eat*100/n)+str(drink*100/n)+str(mem*100/n)+str(lmem*100/n)+str(herd*100/n)+str(forage*100/n)+str(nothing*100/n)+str((int)(energy*100/(n*Sheep.goodFoodLevel)))+str((int)(hydration*100/(n*Sheep.goodWaterLevel)))+"\n",sim.minute());
			System.out.println(""+sim.minute()+" E:"+str(eat*100/n)+" D:"+str(drink*100/n)+" M:"+str(mem*100/n)+" L:"+str(lmem*100/n)+" H:"+str(herd*100/n)+" F:"+str(forage*100/n)+" N:"+str(nothing*100/n)+" EN:"+str((int)(energy*100/(n*Sheep.goodFoodLevel)))+" WA:"+str((int)(hydration*100/(n*Sheep.goodWaterLevel))));
		}
		
		// Remove all animals which died
		if(toDie.size() > 0) removeDead();
	}
	
	public String str(int num){
		if(num < 10) return ""+num+"   ";
		if(num < 100) return ""+num+"  ";
		return ""+num+" ";
	}
	
	/** Returns the animal's physological state */
	public int getState(){ return state; }
	
	/** Calculates the centre of the herd */
	public Point2D.Double calculateHerdCentre(){
		double x = 0;
		double y = 0;
		for(Iterator i = iterator(); i.hasNext();){
			Animal a = (Animal)i.next();
			x += a.x;
			y += a.y;
		}
		return new Point2D.Double(x/(double)n, y/(double)n);
	}


	/** Removes the animal from the list of live animals */
	public void notifyDead(Animal a){ toDie.add(a); }
	
	/** Removes all the dead animals */
	public synchronized void removeDead(){
		for(Iterator i = toDie.iterator(); i.hasNext();){
			Animal a = (Animal)i.next();
			remove(a);
			a.removePos();
			deceased.add(a);
		}
		toDie = new LinkedList();
		n = size();
	}

/*
	This old algorithm checked each animal with every other.
	The new one uses an Animal array to keep track of positions.

	public void disperse(Animal a1){
		Animal a2 = getClose(a1); // Closest animal
		if(a2 != null) a1.separate(a2);
		else{
			for(Iterator j = animals.iterator(); j.hasNext();){
				a2 = ((LivestockGroup)j.next()).getClose(a1);
				if(a2 != null){ a1.separate(a2); }
			}
		}
	}

	
	public Animal getClose(Animal a1){
		for(Iterator j = iterator(); j.hasNext();){
			Animal a2 = (Animal)j.next();
			if(a1.distance(a2) < sim.ANIMAL_SIZE && a1 != a2) return a2;		
		}
		return null;
	}
		
*/
	

	//------------[ Methods for drawing the livetock ]---------------

	public synchronized void drawDead(Graphics g){
		for(Iterator i = deceased.iterator(); i.hasNext();){
			((Animal)i.next()).drawDead(g);
		}
	}
	
	
	public synchronized void draw(Graphics g){
		for(Iterator i = iterator(); i.hasNext();){
			((Animal)i.next()).draw(g);
		}
		// Uncomment to see where they think best pastures
		// are (represented as a purple dot)
		//drawMemory(g);
	}
	
	public synchronized void drawHunger(Graphics g){
		for(Iterator i = iterator(); i.hasNext();){
			((Animal)i.next()).drawHunger(g);
		}
		//drawMemory(g);
	}
	
	public synchronized void drawThirst(Graphics g){
		for(Iterator i = iterator(); i.hasNext();){
			((Animal)i.next()).drawThirst(g);
		}
		//drawMemory(g);
	}
	
	
	public void drawMemory(Graphics g){
		Point2D.Double p = memory.pos();
		g.setColor(Color.MAGENTA);
		sim.fillScaledOval(g,(int)p.x, (int)p.y, (int)sim.METER*4, (int)sim.METER*4);
	}
	
	
	public Simulation getSim(){ return sim; }

	private void writeObject(java.io.ObjectOutputStream out) throws IOException{
		out.writeObject(name);
		out.writeInt(type);
		out.writeInt(state);
		out.writeObject(deceased);
		out.writeObject(memory);
	}
	private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException{
		name = (String)in.readObject();
		type = in.readInt();
		state = in.readInt();
		deceased = (LinkedList)in.readObject();
		memory = (HerdMemory)in.readObject();
	}
	
	// Class representing a herd's memory for good pastures
	class HerdMemory implements Serializable{
		
		private static final int MEMORY_POINTS = 5; // The number of pasture points which contribute to a memory
		public static final int MEMORY_RADIUS = 50; // The radius in meters of the memory distance (affects distribution of memory points)
	
		private LinkedList points;
		private Point2D.Double avgPos;
		private float avgVal;
		private double METER;
		
		// Construct an empty memory.
		public HerdMemory(){
			points = new LinkedList();
			avgPos = new Point2D.Double(-1.0, -1.0);
			avgVal = 0.0f;
			this.METER = sim.METER;
		}
		
		// Returns the average grass value of this memory
		public float value(){ return avgVal; }
		
		// Returns the centre of this memory area
		public Point2D.Double pos(){ return avgPos; }
		
		
		// Notify this memory of a possible pixel to remember
		public void notifyValue(Point2D.Double p, float val){ 
			if(avgPos.x == -1.0){
				avgPos = new Point2D.Double(p.x, p.y);
				points.add(new MemPoint(avgPos, val));
				avgVal = val;
			}
			// If the value is less than the lowest then don't keep it
			if(val <= ((MemPoint)points.getLast()).val && points.size() >= MEMORY_POINTS) return;
			
			// Check if this point is within range
			if(p.distance(avgPos) > MEMORY_RADIUS*METER){
				// If val is grater than gratest then we have a new area
				if(val > ((MemPoint)points.getFirst()).val){
					points = new LinkedList();
					avgPos = new Point2D.Double(p.x, p.y);
					points.add(new MemPoint(avgPos, val));
					avgVal = val;
				}
				return;
			}
			
			// If we reach here then p is within range
			double posX = p.x, posY = p.y; float avg = val; boolean found = false;
			MemPoint mp = null;
			for(ListIterator i = points.listIterator(); i.hasNext();){
				mp = (MemPoint)i.next();
				posX += mp.pos.x; posY += mp.pos.y; avg += mp.val;
				if(!found && val > mp.val){
					i.previous(); // Go back one
					i.add(new MemPoint(new Point2D.Double(p.x, p.y), val)); // Add this new point
					found = true;
					i.next(); // Go forward again
				}
			}
			if(!found) points.add(new MemPoint(new Point2D.Double(p.x, p.y), val)); 
			if(points.size() > MEMORY_POINTS){
				posX -= mp.pos.x; posY -= mp.pos.y; avg -= mp.val;
				points.removeLast(); // Remove the last point
			}
			avgVal = avg/(float)points.size();
			avgPos.x = posX/(double)points.size();
			avgPos.y = posY/(double)points.size();
		}
		
		// Slightly reduces the memory value to ensure that
		// The memory doesn't persist for too long
		public void dampen(){
			for(Iterator i = points.iterator(); i.hasNext();){
				((MemPoint)i.next()).val /= MEM_DAMPEN_FACTOR;
			}
		}
		
		
		private void writeObject(java.io.ObjectOutputStream out) throws IOException{
			out.writeObject(points);
			out.writeDouble(avgPos.x);
			out.writeDouble(avgPos.y);
			out.writeFloat(avgVal);
			out.writeDouble(METER);
		}
		private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException{
			points = (LinkedList)in.readObject();
			avgPos = new Point2D.Double(in.readDouble(), in.readDouble());
			avgVal = in.readFloat();
			METER = in.readDouble();
		}
		
		// Data structure representing a memory point
		class MemPoint implements Serializable{
			public Point2D.Double pos; public float val;
			public MemPoint(Point2D.Double pos, float val){ this.val = val; this.pos = pos; }
			private void writeObject(java.io.ObjectOutputStream out) throws IOException{
				out.writeDouble(pos.x);
				out.writeDouble(pos.y);
				out.writeFloat(val);
			}
			private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException{
				pos = new Point2D.Double(in.readDouble(), in.readDouble());
				val = in.readFloat();
			}
		}
	
	}
}
