/*
 * File:   Sheep.java
 * Author: Conrad Rider
 * Email:  c.rider@eviltree.co.uk
 * Date:   22-02-05
 */


// http://agspsrv34.agric.wa.gov.au/agency/pubns/farmnote/1991/F05491.htm (stocking rates)
// http://www.noble.org/Ag/Livestock/WaterImportantNutrient/ (water)
// http://www.sheepandgoat.com/articles/feedingsmallruminants.html (water)
// http://www.gov.on.ca/OMAFRA/english/engineer/facts/86-053.htm (water)
// http://www.ext.vt.edu/pubs/sheep/410-366/410-366.html#L3 (stocking rate)
 
package sim.livestock;

import sim.*;
import sim.pasture.*;
import sim.livestock.*;
import utils.*;


import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import java.io.*;


/** The sheep, currently the only animal implemented. This
  * class deals specifically with sheep's behaviour */
public class Sheep extends Animal{


	// Sheep constants - feel free to tweak

	public static final float SHEEP_DAILY_FOOD       = 600.0f;  // Dry Matter required in g/day
	public static final int   FOOD_SURVIVAL_DAYS     = 60;       // 60 Number of days the sheep will survive without food
	public static final float SHEEP_CONSUME_RATE     = 4.0f;    // Consumption rate g/minute
	
	public static final float SHEEP_DAILY_WATER      = 4.25f;   // Number of litres of water required daily
	public static final int   WATER_SURVIVAL_DAYS    = 7;       // Number of days a sheep will survive without water
    	public static final float SHEEP_DEHYDRATION_RATE = SHEEP_DAILY_WATER/1440.0f; 
	
	public static final int   MAX_WATER_DIST         = 2000; // [2Km] The number of meters at which the distance of a water source is too far
	
	public static final double SWEEP_ANGLE           = Math.toRadians(20.0); // The angle from left to right in the sheep's eating arc
	
	public static final float AVERAGE_SPEED_CONST    = 33.3f; //number of meters per minute (2Km/h)	
	public float AVERAGE_SPEED; // the actual average speed
	public float GRAZING_SPEED; // the grazing speed
	public float FORAGE_SPEED;  // The foraging speed
	
	public static final int FOLLOW_DISTANCE          = 15; // The max number of meters which an animal will follow from
	public static final int DANGER_DIST              = 10; // Number of meters away from small herd before an animal is in danger, controls herd density
	
	public static final float MEM_ATTRACT_RADIUS = 2.0f; // The distance of attraction to the memorised pasture
	public static final float LMEM_RADIUS = 20.0f;       // Local mem attract radius
	
	public static final float MEMORY_THRESHOLD  = 0.5f;  // If grass value is lower than this the animal will move to remembered pasture (the less energy the lower this threshold)
	public static final float LMEMORY_THRESHOLD = 0.5f;  // If grass value is lower than this the animal will move to the local best remembered pixel
	public static final float FORAGE_THRESHOLD  = 0.5f;  // If grass value is lower than this, but higher than memory the animal will forage where it is for better grasses
	
	public static final float LMEM_DAMPEN_FACTOR = 1.5f; // How quickly it forgets best local pasture
	
	// Used to give the animals an id, for debug purposes
	public static int sheep_count;
	private int id;
	
	// The areas that a sheep may desire
	private static final int DESIRE_NONE    = 0;
	private static final int DESIRE_WATER   = 1;
	private static final int DESIRE_HERD    = 2;
	private static final int DESIRE_SLEEP   = 3;
	private static final int DESIRE_MEMORY  = 4;
	private static final int DESIRE_LMEMORY = 5;
	
	
	private transient Point2D.Double  lastLastPos;     // The last last position of this animal
	private transient Point2D.Double  lastPos;         // The last position of this animal
	private transient Point2D.Double  localHerdCentre; // The centre of our local herd
	private           int             herdCount;       // Counts how long since we last did a herd check
	private           int             desicion;        // The last desicion taken
	public  transient int             simTime;         // The current sim time
	private           int             lastDesicion;    // The last desicion taken
	private transient Sheep           leader;          // The sheep we are currently following
	private           int             desiredArea;     // The area the sheep desires
	public transient Point2D.Double   bestPixel;       // The best pixel so far
	public transient float            bestPixelVal;    // The best value so far
	private           int             rndTime;         // A random time offset
	

/*HACK*/	public transient int followCount; // So they dont 
/*HACK*/	public static final int FOLLOW_TIME = 200;
	
	public static final float goodFoodLevel = SHEEP_DAILY_FOOD*(float)FOOD_SURVIVAL_DAYS;
	public static final float goodWaterLevel = SHEEP_DAILY_WATER*(float)WATER_SURVIVAL_DAYS;
	
	
	public Sheep(Simulation sim, double px, double py, LivestockGroup herd){
		super(sim, px, py, herd);
		
		AVERAGE_SPEED = (float)sim.METER * AVERAGE_SPEED_CONST;
		GRAZING_SPEED = (float)sim.METER * AVERAGE_SPEED_CONST / 33.3f;
		FORAGE_SPEED  = GRAZING_SPEED * 4.0f;
				
		id = sheep_count;
		sheep_count++;
		
		// 60 days worth of energy
		energy = goodFoodLevel + ((float)Math.random()-0.25f)*SHEEP_DAILY_FOOD*2.0f; // Ensure they don't all start the same
		hydration = goodWaterLevel + ((float)Math.random()-0.5f)*SHEEP_DAILY_WATER*2.0f; // Ensure they don't all start the same
		
		// Set random time offset
		rndTime = (int)((Math.random()-0.5)*10.0); //animal will be 5 minutes early/late
		
		leader = null;
		// Randomise so that different sheep scan at different times
		herdCount = (int)(Math.random()*20.0);
		localHerdCentre = new Point2D.Double(x, y);
		desiredArea = DESIRE_NONE;
		alive = true;
	}
	
	
	public int getDesicion(){ return desicion; }

	public void draw(Graphics g){
		g.setColor(Animal.COLOR_SHEEP);
		super.draw(g);
	}
	
	public void drawDead(Graphics g){
		g.setColor(Animal.COLOR_SHEEP);
		super.drawDead(g);
	}
	
	/** Draw hunger data */
	public void drawHunger(Graphics g){
		//drawMemory(g);
		float val = (energy-goodFoodLevel)/SHEEP_DAILY_FOOD;
		      val = (val >= 0.0f)? val : 0.0f;
		      val = (val <= 1.0f)? val : 1.0f;
		g.setColor(new Color(1.0f-val, val, 0.0f));
		sim.fillScaledOval(g, x-sim.ANIMAL_SIZE/2, y-sim.ANIMAL_SIZE/2, sim.ANIMAL_SIZE, sim.ANIMAL_SIZE);
		Point2D.Double head = direction.project(x, y, sim.ANIMAL_SIZE/2.0);
		g.setColor(Color.GREEN);
		sim.fillScaledOval(g, head.x-sim.ANIMAL_SIZE/4.0, head.y-sim.ANIMAL_SIZE/4.0, sim.ANIMAL_SIZE/2.0, sim.ANIMAL_SIZE/2.0);
	}
	
	/** Draw thirst data */
	public void drawThirst(Graphics g){
		//drawMemory(g);
		float val = (hydration-goodWaterLevel)/SHEEP_DAILY_WATER;
		if(val > 1.0f) val = 1.0f;
		if(val < 0.0f) val = 0.0f;
		g.setColor(new Color(1.0f-val, 0.0f, val));
		sim.fillScaledOval(g, x-sim.ANIMAL_SIZE/2, y-sim.ANIMAL_SIZE/2, sim.ANIMAL_SIZE, sim.ANIMAL_SIZE);
		Point2D.Double head = direction.project(x, y, sim.ANIMAL_SIZE/2.0);
		g.setColor(Color.BLUE);
		sim.fillScaledOval(g, head.x-sim.ANIMAL_SIZE/4.0, head.y-sim.ANIMAL_SIZE/4.0, sim.ANIMAL_SIZE/2.0, sim.ANIMAL_SIZE/2.0);
	}
	
	public void place(Point2D.Double p){
		super.place(p);	
		localHerdCentre = new Point2D.Double(p.x, p.y);
	}
	
	
	/** Draw thirst data */
	public void drawMemory(Graphics g){
		if(bestPixel != null){
			g.setColor(Color.CYAN);
			sim.fillScaledOval(g,(int)bestPixel.x, (int)bestPixel.y, 1, 1);
		}
	}
	

	
//--------------------------[ This gives the sheep life ]-----------------------

	/** This gets the sheep to do what its going to do */
	public void animate(){
//System.out.print("("+desiredArea+")");
		super.animate();
		herdCount++;
		
		



		
		//if(desiredArea != DESIRE_NONE){
		simTime = sim.dayMinute();
		
		// Reduce water level (every 4 hours)
		if(simTime % (360+rndTime) == 0) hydration -= SHEEP_DEHYDRATION_RATE*360;
		
		// Reduce local memory
		if(bestPixel != null){
			bestPixelVal = grass.value((int)(bestPixel.x+0.5), (int)(bestPixel.y+0.5), 70)/1500.0f;
			if(bestPixelVal == 0.0f){
				if(desiredArea == DESIRE_LMEMORY) desiredArea = DESIRE_NONE;
				bestPixel = null;
			}
		}///= LMEM_DAMPEN_FACTOR; }
		
		// Reduce energy at 6am to start consumption
		if(simTime == (420+rndTime) || simTime == (900+rndTime)) energy -= SHEEP_DAILY_FOOD/2.0;
		
		// If the animal has run out of fuel it has died
		if(energy <= 0 || hydration <= 0){ herd.notifyDead((Animal)this); alive = false; return; }
		
		// Do memory stuff
//float grassVal = bounds.fullyHolds(this)? grass.value((int)(x+0.5), (int)(y+0.5), 70)/1500.0f : 0.0f;
		float grassVal = grass.value((int)(x+0.5), (int)(y+0.5), 70)/1500.0f;
		// Check if better than best pixel
		if(grassVal > bestPixelVal){
			bestPixel = new Point2D.Double(x, y);
			bestPixelVal = grassVal;
			herd.memory.notifyValue(bestPixel, grassVal);
			
			if(leader != null) leader.notifyBestPixel(bestPixel, bestPixelVal);		
		}
		

/*HACK*///if(desiredArea == DESIRE_MEMORY && simTime > 600){ desiredArea = DESIRE_NONE; leader = null; }
		

		// If the this animal is following another, let it follow
		if(leader != null){ follow(); return; }
// Incrament the allowed follow time
/*HACK*/else followCount = (followCount < FOLLOW_TIME)? followCount++ : followCount;
		
		// If animal already has a desire, satisfy it
		switch(desiredArea){
			case DESIRE_WATER   : drink(); return;
			case DESIRE_MEMORY  : memoryRoam(); return;
			case DESIRE_LMEMORY : localMemoryRoam(); return;
		}
		
		// Calculate where our local herd centre is
		if(herdCount >= 20){ localHerdCentre = findLocalHerd(); herdCount = 0; }
		
		// If not it needs to make a desicion
		desicion = makeDesicion(grassVal);
		switch(desicion){
			case DO_NOTHING  : stand(); break;
			case EAT         : eat(); break;
			case FORAGE      : forage(); break;
			case DRINK       : desiredArea = DESIRE_WATER; drink(); break;
			case MEMORY      : desiredArea = DESIRE_MEMORY; memoryRoam(); break;
			case LMEMORY     : desiredArea = DESIRE_LMEMORY; localMemoryRoam(); break;
			case HERD        : herd(); break;
		}
		
		lastDesicion = desicion;
	}
	
//---------------------[ This is where the animal makes its desicion ]---------------
	
	
	/** This is where the animal decides what to do */
	public int makeDesicion(float grassVal){
		float m, m_;
		      grassVal = (grassVal >= 0.0f)? grassVal : 0.0f;
		      grassVal = (grassVal <= 1.0f)? grassVal : 1.0f;
		float energyVal = (energy-goodFoodLevel)/(SHEEP_DAILY_FOOD*2.0f);
		      energyVal = (energyVal >= 0.0f)? energyVal : 0.0f;
		      energyVal = (energyVal <= 1.0f)? energyVal : 1.0f;
		int desicion = DO_NOTHING;
		// If there is no other motive grater than 0.01
		// then the animal won't bother doing anything 
		m = (simTime > (360+rndTime) && simTime < (1080+rndTime))? 0.1f : 1.0f; // Sleeping between 18:00 and 06:00 
		m_ = getSafetyMotive();         if(m_ > m){ m = m_; desicion = HERD; }
		m_ = getDrinkMotive();          if(m_ > m){ m = m_; desicion = DRINK; }
		// Make desicion about eating. IF energy is high there will
		// be more motive to search and forage, if it is low then
		// the animal will just eat what it can
		if(1.0-energyVal > m){
			
			grassVal = (float)Math.pow(grassVal, 2);
			
			if     (grassVal <= MEMORY_THRESHOLD * energyVal &&
			        (simTime == (450+rndTime) || simTime == (870+rndTime)) && //Herd at 7:30 and 2:30  
			        this.distance(herd.memory.pos()) >
			        Math.sqrt(herd.getSize())*sim.METER*MEM_ATTRACT_RADIUS) desicion = MEMORY;
			else if(grassVal < LMEMORY_THRESHOLD * energyVal &&
			        simTime % 20 == 0 &&
				bestPixel != null &&
				this.distance(bestPixel) >
				sim.METER*LMEM_RADIUS) desicion = LMEMORY;
			else if(grassVal < FORAGE_THRESHOLD * energyVal) desicion = FORAGE;
			else{
				if(grassVal > 0.0) desicion = EAT;
				else{
					if(Math.random() < 0.9) desicion = FORAGE;
					else desicion = (simTime == (450+rndTime))? MEMORY : LMEMORY; 
				}
//desicion = EAT; 
			}
		}
//if(m > 0.99 && desicion == EAT){ System.out.println("energy"+energyVal+" m = "+m); }
//if(energyVal <= 0.0f) System.out.println("energy = 0");
		//m_ = getForageMotive(grassVal, energyVal); if(m_ > m){ m = m_; desicion = FORAGE; }
		//m_ = getEatMotive(grassVal, energyVal);    if(m_ > m){ m = m_; desicion = EAT; }
		//m_ = getMemoryMotive(grassVal); if(m_ > m){ m = m_; desicion = MEMORY; }
//System.out.println("Eat motive = "+m_);
//if(desicion == DO_NOTHING) System.out.println("\nGrass = "+grassVal+" Energy = "+energyVal+" m_ = "+m_);
//if(energyVal > 0.99) System.out.println("\nenergy = "+energyVal);
		
		return desicion;
	}

	
	/** Returns the amount of motive the sheep has to eat.
	  * Motive varies inversly with energy and
	  * directly with quality of grass it is standing at */
//	private float getEatMotive(float grass, float energy){ return (energy < 1.0f)? (1.0f - energy) : 0.0f; }
	
	
	/** Returns the motive to search for grass other than that which is at its feet */
//	private float getForageMotive(float grass, float energy){ return (energy < 1.0f)? (1.0f - (float)Math.pow(grass, 0.5)) : 0.0f; }

	
	/** Returns the motive to go to last good pasture area */
//	private float getMemoryMotive(float grass){ return (grass < 0.9f)? 0.11f : 0.0f; }
	
		
	/** Returns the level of motive the sheep has to drink */
	private float getDrinkMotive(){
		// Find closest water area
		float waterDistance;
		waterDistance = sim.getWaterPath().getDistance((int)(x+0.5),(int)(y+0.5));
		
		// There is no motive to move towards non-existant water
		if(waterDistance < 0.0f) return 0.0f;
		
		// The sheep will not go to water which is too far away
		if(waterDistance > MAX_WATER_DIST) return 0.0f;
		
		// Convert distance into a number between 0 and 1
//		waterDistance = waterDistance/(float)MAX_WATER_DIST;
//		if(waterDistance > 1.0f) waterDistance = 1.0f;
		
		// Cap the water distance to prevent motive from going to 0
//		if(waterDistance < 0.2f){ waterDistance = 0.2f; }

		// Motive is inversly proportional to level of hydration
		float val = 1.0f-(hydration-goodWaterLevel)/(SHEEP_DAILY_WATER*2.0f);
		return (val < 0.99)? val : 1.0f;
	}
	
	
	/** Returns the amount of motive the sheep has to herd */
	private float getSafetyMotive(){
		
		if(localHerdCentre == null) return 0.0f;
		float d = (float)this.distance(localHerdCentre)/(float)(DANGER_DIST*sim.METER);
		// Prevent them from going too far in search of grass
		if(d > 1.0f){ d = 1.0f; }
		// Prevent them from herding at close distance
		if(d < 0.5) d = 0.0f;
		return d;
	}


	
// ------------[ This is where the sheep reacts to a desicion ]----------------
	
	/** Make the sheep eat */
	public void eat(){
//System.out.print("e");
		
		// Create all directions to scan in
		Vector2D[] D = {
			direction,
			direction.rotate(2*Math.PI-SWEEP_ANGLE),
			direction.rotate(SWEEP_ANGLE),
			direction.rotate(2*Math.PI-SWEEP_ANGLE*2),
			direction.rotate(SWEEP_ANGLE*2),
			//direction.rotate(2*Math.PI-SWEEP_ANGLE*3),
			//direction.rotate(SWEEP_ANGLE*3)
		};
		// Project to get points and then find value
		Point2D.Double p = D[1].project(x, y, GRAZING_SPEED);
		float v = grass.value((int)(p.x+0.5), (int)(p.y+0.5), 70);
		for(int i = 1; i < D.length; i++){
			p = D[i].project(x, y, GRAZING_SPEED);
			float v_ = grass.value((int)(p.x+0.5), (int)(p.y+0.5), 70);
			if(v_ > v){ v = v_; direction = D[i]; }
		}
				
		// Consume the grass (numbers from http://www.iger.bbsrc.ac.uk/Publications/Innovations/in97/Ch9.pdf)
		energy += grass.visit((int)(x+0.5), (int)(y+0.5), PastureState.VISIT_GRAZE, SHEEP_CONSUME_RATE, 70);

		roam(direction, GRAZING_SPEED); // Animal creeps forward as it grazes
	}
	
	
	/** Make the sheep forage for grass */
	public void forage(){
//System.out.print("f");
		
		// Create all directions to forage in
		Vector2D[] D = {
			direction,
			direction.rotate(2*Math.PI-SWEEP_ANGLE),
			direction.rotate(SWEEP_ANGLE),
			direction.rotate(2*Math.PI-SWEEP_ANGLE*2),
			direction.rotate(SWEEP_ANGLE*2),
			direction.rotate(2*Math.PI-SWEEP_ANGLE*3),
			direction.rotate(SWEEP_ANGLE*3)
		};
		// Project to get points and then find value
		Point2D.Double p = D[1].project(x, y, FORAGE_SPEED);
		int v = (int)(grass.avgVal((int)(p.x+0.5), (int)(p.y+0.5), 0));
		for(int i = 1; i < D.length; i++){
			p = D[i].project(x, y, FORAGE_SPEED);
			int v_ = (int)(grass.avgVal((int)(p.x+0.5), (int)(p.y+0.5), 0));
			if(v_ > v){ v = v_; direction = D[i]; }
		}
		
		// Walk over the patch
		grass.visit((int)(x+0.5), (int)(y+0.5), PastureState.VISIT_WALK);
		
		roam(direction, FORAGE_SPEED); // Animal creeps forward as it grazes
	}
	
	
	/** Make the sheep seek or drink water */
	private void drink(){
//System.out.print("d");
		// Find the direction of the water
		//Vector2D waterDirection = new Vector2D(x,y,closestWater.x, closestWater.y);
		PathFinderMap map = sim.getWaterPath();
		Vector2D waterDirection = map.getDirection((int)(x+0.5),(int)(y+0.5));
		Point2D.Double wp = waterDirection.project(x,y,1.4);
		//if(sim.getWaterGeometry().inWater(wp.x, wp.y)){
		float dist = map.getDistance((int)(wp.x), (int)(wp.y));
		if(dist == 0.0f){
			grass.visit((int)(x+0.5), (int)(y+0.5), PastureState.VISIT_STAND);
			hydration += SHEEP_DAILY_WATER*4;
			super.roam((Point2D.Double)this); // This just ensures sheap avoid collision.
			desiredArea = DESIRE_NONE; // We have found the water, so no more desire to find it
		}
		else if(dist > 0.0f && dist <= FORAGE_SPEED){
			direction = gravitatePaths(waterDirection);
			roam(direction, (dist+1.0f)/2.0f);
		}
		else{
			grass.visit((int)(x+0.5), (int)(y+0.5), PastureState.VISIT_WALK);
			if(findLeader(DESIRE_WATER)){
				// Move towards the leader
				Vector2D d = handleGates(leader);
				if(d == null){ // Means the position is unreachable
					direction = waterDirection;
					leader = null;
				}else direction = d;
			}
			else{
				// Move towards the water
				direction = waterDirection;
				leader = null;
			}
			direction = gravitatePaths(direction);
			roam(direction, FORAGE_SPEED);
		}
	}

	
	/** Make the sheep move towards the herd */
	public void herd(){
//System.out.print("h");
		grass.visit((int)(x+0.5), (int)(y+0.5), PastureState.VISIT_WALK);
		
		// Move towards herd centre
		Vector2D d = handleGates(localHerdCentre);
		if(d == null) direction = new Vector2D();
		else direction = d;
		
		roam(direction, FORAGE_SPEED); // Roam in the right direction
	}
	
	
	
	/** Make the sheep roam towards better pastures (remembered by self)*/
	private void localMemoryRoam(){
//System.out.print("m");
		if(bestPixel == null){ desiredArea = DESIRE_NONE; forage(); return; }
		if(this.distance(bestPixel) < sim.METER*LMEM_RADIUS-(Math.random()*sim.METER*LMEM_RADIUS*0.5)){
			grass.visit((int)(x+0.5), (int)(y+0.5), PastureState.VISIT_STAND);
			super.roam((Point2D.Double)this); // This just ensures sheap avoid collision.
			desiredArea = DESIRE_NONE; // We have found the good area, so no more desire to find it
		}
		else{
			grass.visit((int)(x+0.5), (int)(y+0.5), PastureState.VISIT_WALK);
			if(findLeader(DESIRE_LMEMORY) && !bounds.onGate(this)){
//System.out.println("Found mem leader");
				// Move towards the leader
				Vector2D d = handleGates(leader);
				if(d == null){ // Means the position is unreachable
					d = handleGates(bestPixel);
					if(d == null){ // Means the position is unreachable
						desiredArea = DESIRE_NONE;
						direction = new Vector2D();
					}else direction = d; 
					leader = null;
				}else direction = d;
			}
			else{
// If best pixel is too low then don't bother
//if(bestPixelVal < 0.5){ desiredArea = DESIRE_NONE; forage(); return; }
				// Move towards the best grass directly
				Vector2D d = handleGates(bestPixel);
				if(d == null){ // Means the position is unreachable
					desiredArea = DESIRE_NONE;
					direction = new Vector2D();
				}else direction = d;
				
				leader = null;
			}
			direction = gravitatePaths(direction);
			roam(direction, FORAGE_SPEED);
		}
	}
	

	
	/** Make the sheep roam towards better pastures (remembered by herd)*/
	private void memoryRoam(){
//System.out.print("m");
		
		if(this.distance(herd.memory.pos()) < Math.sqrt(herd.getSize())*sim.METER*MEM_ATTRACT_RADIUS*Math.random()){
			grass.visit((int)(x+0.5), (int)(y+0.5), PastureState.VISIT_STAND);
			super.roam((Point2D.Double)this); // This just ensures sheap avoid collision.
			desiredArea = DESIRE_NONE; // We have found the good area, so no more desire to find it
		}
		else{
			grass.visit((int)(x+0.5), (int)(y+0.5), PastureState.VISIT_WALK);
			if(findLeader(DESIRE_MEMORY)){
				// Move towards the leader
				Vector2D d = handleGates(leader);
				if(d == null){ // Means the position is unreachable
					d = handleGates(herd.memory.pos());
					if(d == null){ // Means the position is unreachable
						desiredArea = DESIRE_NONE;
						direction = new Vector2D();
					}else direction = d; 
					leader = null;
				}else direction = d;
			}
			else{
				// Move towards the best grass directly
				Vector2D d = handleGates(herd.memory.pos());
				if(d == null){ // Means the position is unreachable
					desiredArea = DESIRE_NONE;
					direction = new Vector2D();
				}else direction = d; 
				
				leader = null;
			}
			direction = gravitatePaths(direction);
			roam(direction, FORAGE_SPEED);
		}
	}
	

	/** Follow an animal */
	public void follow(){
// Decrament follow count since we are following
/*HACK*/followCount -= 2;
/*HACK*/		if(followCount < -FOLLOW_TIME){ desiredArea = DESIRE_NONE; leader = null; return; }
//System.out.print("f");
		// Check if the leader is still suitable
		if(leader.lookingFor(desiredArea)){
			grass.visit((int)(x+0.5), (int)(y+0.5), PastureState.VISIT_WALK);
			Vector2D d = handleGates(leader);
			if(d == null){
				direction = new Vector2D(); leader = null;
			}
			else direction = d;
			double dist = this.distance(leader);
			dist = (dist/2.0 < AVERAGE_SPEED)? dist/2.0 : AVERAGE_SPEED;
			roam(direction, (float)dist);
		}
		// If leader is doing something else then use initiative
		else{
			leader = null;
			direction = new Vector2D(0, 0);
			if(desiredArea == DESIRE_WATER) drink();
			if(desiredArea == DESIRE_MEMORY) memoryRoam();
			if(desiredArea == DESIRE_LMEMORY) localMemoryRoam();
		}
	}
	
	
	/** The sheep just stands there looking stupid! */
	private void stand(){
//System.out.print("n");
		grass.visit((int)(x+0.5), (int)(y+0.5), PastureState.VISIT_STAND);
		// Do not allow the animal to breach others' space or field bounds
		super.roam((Point2D.Double)this);
	}
	
	
	/** Prevents the animal from moving directly towards objects in
	  * different fields. Insted it will find a gate which leads to
	  * the correct field */
	private Vector2D handleGates(Point2D.Double d){ 
		// We do not have to pass any gates
		if(bounds.holds(d)) return new Vector2D(this, d); 
		
		
		//Do dijkstra's variant to find shortest path to d
		FieldGate g = bounds.findGate(d);
		
//if(g == null){
//	System.out.println("Going from "+bounds+" to "+newBounds+" via "+g);
//	System.out.println("bounds holds?"+ bounds.holds(d) + " newbounds?"+newBounds.holds(d)+" point="+d);
//}
		
		// If the result is null then just return null
		if(g == null) return null;
		
		// If we are over the gate then move towards the field we are looking for
		//if(this.distance(g) < sim.GATE_WIDTH*0.5){  System.out.println("H case 5");  return g.towards(newBounds); }
		
		// If not then return direction to found gate
		return new Vector2D((Point2D.Double)this, (Point2D.Double)g);
	}
	
		
	/** Roam in a random direction */
	public void roam(){ roam(new Vector2D(), AVERAGE_SPEED); }
	// Roam in a particular direction at average speed
	public void roam(Vector2D v){ roam(v,AVERAGE_SPEED); }
	// Roam in a particular direction at particular speed
	public void roam(Vector2D v, float speed){
		if(stuck()){
			v = new Vector2D();
			while(v.isZeroVector()){ direction = new Vector2D(); }
			speed = FORAGE_SPEED;
		}
		lastLastPos = lastPos;
		lastPos = new Point2D.Double(x,y);
		// Take of the energy it requires
		//energy -= Math.ceil(speed * MOVEMENT_NUTRITION);
		Point2D.Double next = v.project(x,y, speed);
		super.roam(next);
	}

	/* Prevents the animal from getting stuck against an object */
	public boolean stuck(){
//		if(lastPos != null && this.distance(lastPos) < (GRAZING_SPEED*0.001)){
//			System.out.print(" LOW:"+desicion);
//		}
		return (lastDesicion == desicion &&
		        (lastPos != null &&
			 this.distance(lastPos) == 0.0)      );/* ||
		        (lastLastPos != null &&
   		         this.distance(lastLastPos) == 0.0)); */
	}
	
	
	
// ----------------------------- Following algorithm ----------------
	
		
	/** Scan for an animal looking for seekObj to follow */
	public boolean findLeader(int seekObj){
		// Using ceiling to ensure that animals will definetly move initially
		int scanArea = (int)sim.METER*FOLLOW_DISTANCE;
		int x = (int)this.x;
		int y = (int)this.y;
		
		if(withinArea(x, y)
		   && pos[x][y] != this
		   && pos[x][y] != null){
			if(suitableLeader(pos[x][y], seekObj)){
				leader = (Sheep)pos[x][y];
				return true;
			}
		}
			
		// Scans for the closest animal which is closer than scanArea
		for(int i = 1; i <= scanArea; i++){
			
			Vector cand = new Vector();
			if(withinArea(x+i,y) && pos[x+i][y] != null && pos[x+i][y] != this) cand.add(pos[x+i][y]);
			if(withinArea(x-i,y) && pos[x-i][y] != null && pos[x-i][y] != this) cand.add(pos[x-i][y]);
			if(withinArea(x,y+i) && pos[x][y+i] != null && pos[x][y+i] != this) cand.add(pos[x][y+i]);
			if(withinArea(x,y-i) && pos[x][y-i] != null && pos[x][y-i] != this) cand.add(pos[x][y-i]);
			if(leaderFound(cand, seekObj)) return true;
			
			for(int j = 1; j < (int)Math.floor(i/2); j++){
				
				if(distance(new Point2D.Double(x+j, y+i)) < scanArea){
					cand = new Vector();
					if(withinArea(x+j,y+i) && pos[x+j][y+i] != null && pos[x+j][y+i] != this) cand.add(pos[x+j][y+i]);
					if(withinArea(x+j,y-i) && pos[x+j][y-i] != null && pos[x+j][y-i] != this) cand.add(pos[x+j][y-i]);
					if(withinArea(x-j,y+i) && pos[x-j][y+i] != null && pos[x-j][y+i] != this) cand.add(pos[x-j][y+i]);
					if(withinArea(x-j,y-i) && pos[x-j][y-i] != null && pos[x-j][y-i] != this) cand.add(pos[x-j][y-i]);
					if(withinArea(x+i,y+j) && pos[x+i][y+j] != null && pos[x+i][y+j] != this) cand.add(pos[x+i][y+j]);
					if(withinArea(x+i,y-j) && pos[x+i][y-j] != null && pos[x+i][y-j] != this) cand.add(pos[x+i][y-j]);
					if(withinArea(x-i,y+j) && pos[x-i][y+j] != null && pos[x-i][y+j] != this) cand.add(pos[x-i][y+j]);
					if(withinArea(x-i,y-j) && pos[x-i][y-j] != null && pos[x-i][y-j] != this) cand.add(pos[x-i][y-j]);
					if(leaderFound(cand, seekObj)) return true;
				}
			}
			
			if(distance(new Point2D.Double(x+i, y+i)) < scanArea){
				cand = new Vector();
				if(withinArea(x+i,y+i) && pos[x+i][y+i] != null && pos[x+i][y+i] != this) cand.add(pos[x+i][y+i]);
				if(withinArea(x+i,y-i) && pos[x+i][y-i] != null && pos[x+i][y-i] != this) cand.add(pos[x+i][y-i]);
				if(withinArea(x-i,y+i) && pos[x-i][y+i] != null && pos[x-i][y+i] != this) cand.add(pos[x-i][y+i]);
				if(withinArea(x-i,y-i) && pos[x-i][y-i] != null && pos[x-i][y-i] != this) cand.add(pos[x-i][y-i]);
				if(leaderFound(cand, seekObj)) return true;
			}
		}
		return false;
	}
	
	/** Finds the closest animal in the vector to use as leader */
	public boolean leaderFound(Vector v, int seekObj){
		if(v.size() == 0) return false;
		Animal choice = (Animal)v.get((int)(Math.random()*v.size()));
		while(!suitableLeader(pos[(int)choice.x][(int)choice.y], seekObj)){
			v.remove(choice);
			if(v.size() > 0) choice = (Animal)v.get((int)(Math.random()*v.size()));
			else return false;
		}
		if(suitableLeader(pos[(int)choice.x][(int)choice.y], seekObj)){
			leader = (Sheep)pos[(int)choice.x][(int)choice.y];
			return true;
		}
		return false;
	}
	
	/** Returns wheather the animal is looking for that object */
	public boolean suitableLeader(Animal an, int seekObj){
	
// Only allow the animal to follow if it has not been hoaring its follow behaviour
/*HACK*/if(followCount < 0){ return false; }
	
		if(an instanceof Sheep){
			Sheep a = (Sheep)an;
		
			// Check their best grass value
			boolean theirValBetter = false;
			Point2D.Double ap = a.bestPixel;
			if(ap != null){
				float theirVal = grass.value((int)(ap.x+0.5), (int)(ap.y+0.5), 70)/1500.0f;
				if(theirVal >= bestPixelVal && theirVal > 0.0f){
					bestPixelVal = theirVal;
					bestPixel = ap;
					theirValBetter = true;
				}
			}

			// Only follow other sheep looking for the same thing
			if(a.lookingFor(seekObj) &&
			a != this && a.isAlive() && // The sheep should be alive
			a.front() != this // Our leader should not be following us
			){
				// Only follow sheep which are closer to goal
				switch(desiredArea){
					case DESIRE_WATER:
						PathFinderMap map = sim.getWaterPath();
						// Our leader should not be following us
						if(map.getDistance(a.x, a.y) < map.getDistance(x, y)){
							return true; 
						}
						return false;
					case DESIRE_MEMORY:
						// Our leader should not be following us
						if(herd == a.getHerd() && a.distance(herd.memory.pos()) < this.distance(herd.memory.pos())){
	//System.out.println(""+id+" folllowing "+a);
							return true; 
						}
						return false;
					case DESIRE_LMEMORY:
						/*if(simTime == ((Sheep)a).simTime){
							if(((Sheep)a).bestPixelVal > bestPixelVal){
								bestPixelVal = ((Sheep)a).bestPixelVal;
								bestPixel = ((Sheep)a).bestPixel;
								return true; 
							}
						}else{*/
						if(bestPixel.distance(a) < bestPixel.distance(this))
							return theirValBetter;
						//}
						//return false;
				}
			}
		}
//System.out.println("test looking for simobj = "+((Sheep)a).lookingFor(seekObj));
		return false;
	}
	
	// Gives the next estimated position of the animal
	public Point2D.Double nextPos(){ return direction.project(this, 1); }
	
	
	/** Returns wheather this sheep is looking for the seekObj */
	public boolean lookingFor(int seekObj){ return seekObj == desiredArea; }
	
	
	/** Returns the front sheep s*/
	public Sheep front(){
		if(leader == null) return this;
		return leader.front();
	}
	
	public void notifyBestPixel(Point2D.Double p, float val){
		if(val > bestPixelVal){
			bestPixelVal = val;
			bestPixel = p;
			if(leader != null) leader.notifyBestPixel(p, val);
		}
	}
	
	public int getId(){ return id; }
	public boolean equals(Sheep s){ return s.getId() == id; }
	


//------------------------[ Herding algorithm ]-----------------------

	// Scan for all sheep within the radius of DANGER_DIST*2
	public Point2D.Double findLocalHerd(){
		
		int scanArea = (int)sim.METER*DANGER_DIST*2;
		
		int x = (int)this.x;
		int y = (int)this.y;
		
		LinkedList localSheep = new LinkedList();
		
		if(withinArea(x, y)
		   && pos[x][y] != this
		   && pos[x][y] != null){
			return new Point2D.Double(x,y);
		}
			
		// Scans for the closest animal which is closer than scanArea
		for(int i = 1; i <= scanArea; i++){
			
			Vector cand = new Vector();
			if(withinArea(x+i,y) && pos[x+i][y] != null && pos[x+i][y] != this) cand.add(pos[x+i][y]);
			if(withinArea(x-i,y) && pos[x-i][y] != null && pos[x-i][y] != this) cand.add(pos[x-i][y]);
			if(withinArea(x,y+i) && pos[x][y+i] != null && pos[x][y+i] != this) cand.add(pos[x][y+i]);
			if(withinArea(x,y-i) && pos[x][y-i] != null && pos[x][y-i] != this) cand.add(pos[x][y-i]);
			if(cand.size() > 0) return findCentre(cand);
			
			for(int j = 1; j < (int)Math.floor(i/2); j++){
				
				if(distance(new Point2D.Double(x+j, y+i)) < scanArea){
					cand = new Vector();
					if(withinArea(x+j,y+i) && pos[x+j][y+i] != null && pos[x+j][y+i] != this) cand.add(pos[x+j][y+i]);
					if(withinArea(x+j,y-i) && pos[x+j][y-i] != null && pos[x+j][y-i] != this) cand.add(pos[x+j][y-i]);
					if(withinArea(x-j,y+i) && pos[x-j][y+i] != null && pos[x-j][y+i] != this) cand.add(pos[x-j][y+i]);
					if(withinArea(x-j,y-i) && pos[x-j][y-i] != null && pos[x-j][y-i] != this) cand.add(pos[x-j][y-i]);
					if(withinArea(x+i,y+j) && pos[x+i][y+j] != null && pos[x+i][y+j] != this) cand.add(pos[x+i][y+j]);
					if(withinArea(x+i,y-j) && pos[x+i][y-j] != null && pos[x+i][y-j] != this) cand.add(pos[x+i][y-j]);
					if(withinArea(x-i,y+j) && pos[x-i][y+j] != null && pos[x-i][y+j] != this) cand.add(pos[x-i][y+j]);
					if(withinArea(x-i,y-j) && pos[x-i][y-j] != null && pos[x-i][y-j] != this) cand.add(pos[x-i][y-j]);
					if(cand.size() > 0) return findCentre(cand);
				}
			}
			
			if(distance(new Point2D.Double(x+i, y+i)) < scanArea){
				cand = new Vector();
				if(withinArea(x+i,y+i) && pos[x+i][y+i] != null && pos[x+i][y+i] != this) cand.add(pos[x+i][y+i]);
				if(withinArea(x+i,y-i) && pos[x+i][y-i] != null && pos[x+i][y-i] != this) cand.add(pos[x+i][y-i]);
				if(withinArea(x-i,y+i) && pos[x-i][y+i] != null && pos[x-i][y+i] != this) cand.add(pos[x-i][y+i]);
				if(withinArea(x-i,y-i) && pos[x-i][y-i] != null && pos[x-i][y-i] != this) cand.add(pos[x-i][y-i]);
				if(cand.size() > 0) return findCentre(cand);
			}
		}
		// If we fail to find any sheep then just return null
		return null;
	}

	// Finds the centre, even if several sheep have been found
	private Point2D.Double findCentre(Vector cand){
	
		double hx = 0;
		double hy = 0;
		for(Iterator i = cand.iterator(); i.hasNext();){
			Animal a = (Animal)i.next();
			hx += a.x;
			hy += a.y;
		}	
		return new Point2D.Double(hx/cand.size(), hy/cand.size());
	}

	
	public LivestockGroup getHerd(){ return herd; }

	
//------------------[ Gravitating to paths ]----------------------

	/** Causes the animals to gravitate towards paths by
	  * rotating their direction towards a path */
	public Vector2D gravitatePaths(Vector2D direction){
		
		// Get directions
		Vector2D left = direction.rotate(2*Math.PI-SWEEP_ANGLE/2);
		Vector2D right = direction.rotate(SWEEP_ANGLE/2);
		
		// Project to next possible position
		Point2D.Double gLeft = left.project(x, y, AVERAGE_SPEED);
		Point2D.Double gRight = right.project(x, y, AVERAGE_SPEED);
		Point2D.Double gForward = direction.project(x, y, AVERAGE_SPEED);
		
		// find visit frequencies
		int leftVal = (int)(grass.freq((int)gLeft.x, (int)gLeft.y));
		int rightVal = (int)(grass.freq((int)gRight.x, (int)gRight.y));
		int forwardVal = (int)(grass.freq((int)gForward.x, (int)gForward.y));
		
		if(leftVal > forwardVal && leftVal > rightVal) return left;
		if(rightVal > forwardVal && rightVal > leftVal) return right; 		
		
		// Continue forward
		return direction;
	}
			
		
//--------------------[ Debug ]-----------------

    	/** Returns the string version of this sheep */
	public String toString(){
		return "Sheep ["+id+"]";
	}

}
