/*
 * File:   GrassModel.java
 * Author: Conrad Rider
 * Email:  c.rider@eviltree.co.uk
 * Date:   22-02-05
 */

package sim.model;


import sim.Simulation;
import java.io.Serializable;


/** This models the pasture growth of each pixel. It is
  * based on the model used in the SEPATOU system */
public class GrassModel implements Serializable{

	// Grass species
	public static final int COCKSFOOT           = 0;
	public static final int PERRENIAL_RYE_GRASS = 1;
	public static final int TALL_FESCUE         = 2;
	
	// The number of weeks to grow the grass for
	public static final int PREPARATION_WEEKS = 0;//4;
	
	// Species dependent constants
	private double K_sward;       // Constant which characterises the vertical structure of the grass
	private int D_max;            // Maximum leaf digestibility
	private int LLS_min, LLS_max; // Max and min leaf lifespan

	private transient double AW;  // Available water (mm)
	private transient double NI;  // Nitrogen index (0-100)
	private transient double LAI; // Leaf area index (0-15)
	private transient double DM;  // Herbage dry matter per m^2
	private transient double SWC; // Soil Water Capacity (0-200) mm (see http://www.ce.utexas.edu/prof/maidment/atlas/world.htm)

	public double Rain = 2.4;     // Rainfall (mm)
	public double T = 10;         // Temperature (celcius) (see http://www.metoffice.com/research/hadleycentre/CR_data/Monthly/HadCET_act_graph.gif)
	public double RG = 25.0;      // Incident photosynthetic radiation MJ.m^-2 (http://www.apogee-inst.com/faq_solar.htm)
	public double PET = 2.73;     // Potential evapo transpiration (0-250) mm (see http://www-cger.nies.go.jp/grid-e/gridtxt/pet.html)

	private transient Simulation sim; // A reference to the simulation
	
	public double getRainfall()     { return Rain; }
	public double getTemperature()  { return T;    }
	public double getRadiation()    { return RG;   }
	public double getTranspiration(){ return PET;  }
	public void setRainfall(double r)     { this.Rain = r; }
	public void setTemperature(double t)  { this.T = t;     }
	public void setRadiation(double r)    { this.RG = r;    }
	public void setTranspiration(double t){ this.PET = t;   }
	
	/** Constructor */
	public GrassModel(Simulation sim){ this(sim, COCKSFOOT); }

	/** Constructor */
	public GrassModel(Simulation sim, int species){

		 this.sim = sim;
		 
		// Initialise constants
		switch(species){
			case COCKSFOOT:
				K_sward = 1.0;
				D_max = 90;
				LLS_min = 600; LLS_max = 800;
				break;
			case PERRENIAL_RYE_GRASS:
				K_sward = 1.2;
				D_max = 95;
				LLS_min = 500; LLS_max = 700;
				break;
			case TALL_FESCUE:
				K_sward = 1.1;
				D_max = 85;
				LLS_min = 700; LLS_max = 900;
				break;
		}
	}
	
	public void initialise(Simulation sim){ this.sim = sim; }


	// Initialise variables
	public void initVars(){

		// Water variables
		AW = 50; // Available water (0-250) mm

		// Soil variables
		NI = 0; // Nitrogen index (0-100)

		// Plant variables
		LAI = 1.8; // Leaf area index (0-15)
		DM = 0; // Dry herbage matter (0-2000) 0.1g.m^-2
	}


	/** Set model data from bit stuffed pixels */
	public void setData(int NI, int SWC, int AW, int LAI, int DM){
		
		// NI  = 4bits  (0-15)   : domain 0-100  mm
		// SWC = 4bits  (0-15)   : domain 0-200  mm
		// AW  = 8bits  (0-255)  : domain 0-200  mm
		// LAI = 12bits (0-4095) : domain 0-6    m/m
		// DM  = 12bits (0-4095) : domain 0-1500 g/m *10

		this.NI  = (double)NI  * 100.0 /  15.0;
		this.SWC = (double)SWC * 200.0 /  15.0;
		this.AW  = (double)AW  * 200.0 / 255.0;
		this.LAI = (double)LAI *   6.0 /4095.0;
		this.DM  = (double)DM  *1500.0 /4095.0;
	}


	/** Return field values in correct domains for available bits */
	public int getNI() { return (int)((NI  *  15.0 / 100.0)+0.5); }
	public int getSWC(){ return (int)((SWC *  15.0 / 200.0)+0.5); }
	public int getAW() { return (AW  >= 0)? ((AW  <=  200)? (int)((AW  * 255.0 / 200.0)+0.5) : 255)  : 0; }
	public int getLAI(){ return (LAI >= 0)? ((LAI <=    6)? (int)((LAI *4095.0 /   6.0)+0.5) : 4095) : 0; }
	public int getDM() { return (DM  >= 0)? ((DM  <= 1500)? (int)((DM  *4095.0 /1500.0)+0.5) : 4095) : 0; }


	/** Computes data for next day */
	public void simulate(){

		// Model the change in water (the water deficit index, WI)
		double AET = Math.min(PET, AW); // Current evapo transpiratoin
		double WI = AET/PET; // Current water deficit index
		AW = Math.min(Rain + AW - AET, SWC); // Compute next available water


		// Model the change in DM (Dry Matter)

		// DRY MATTER INCREASE
		double _DM = DM; // Save old DM
		double ad = 0.1; // Ahem... value was made up, I think it should be 1 if left constnat but I "udjusted" it get sensible output
		double al = 2.5; //double al = -(0.6 / 180) * d + 2.5 + 32 * (0.6 / 180);
		double b = 0.037 + 0.09 * Math.min(T, 18) - 0.0022 * Math.pow(Math.min(T, 18), 2);
		// Intercepted photosynthetic radiation
		double PAR = 0.48 * (0.95 * (1 - Math.pow(Math.E, (-0.54 * LAI)))) * RG;
		// Ratidation use efficiency
		double RUE = ad * al * b * 2.2 * NI * WI;
		double DMr = RUE * PAR;

		// DRY MATTER DECREASE
		double DMs = 0.85 * DM * (Math.min(T, 18)/LLS_min);
		DM += DMr - DMs;
		// Less than 50g.m^-2 is ungrazable by sheep DM of 5


		// Model change in LAI (Leaf Area Index)
		double LAIr = Math.max(0.000011 * Math.pow(Math.min(T, 18), 2) * (NI-20) * WI,0);
		double LAIs = LAI * (Math.min(T, 18)/LLS_min);
		LAI += LAIr - LAIs;
		
		//LAI = K_sward * 1.9 * Math.pow(0.01 * DM, 0.73); // LAI value at the end of winter
		//LAI = (K_sward * 1.9 * Math.pow(0.01 * DM, 0.73)) * (DM/_DM); // LAI value following a defoliation
	}
	
	
	/** Simualtes the growth of pastures for 5 years to initialise the variables */
	public void prepare(){
		// Grow the pasture in preparation for grazing
		for(int w=0; w < PREPARATION_WEEKS; w++){
			for(int d=0; d < 7; d++){ simulate(); }
		}
	}

	
	public float factorReal(float n)  { return n/(float)sim.METER_SQUARED;	}
	public float factorScaled(float n){ return n*(float)sim.METER_SQUARED;	}
	
	// Returns the resulting dry herbage matter when grass (in grams) is consumed
	public float consume(float mass, int height){
		// Scale mass to meter squared
		mass = factorScaled(mass);
		float res = 0.0f;
		double _DM = DM; 
		double _LAI = LAI;
		if(DM > mass*10){
			res = (float)mass*(float)LAI/3.0f;
			DM -= mass*10;
			LAI -= (K_sward * 1.9 * Math.pow(0.01 * DM, 0.73)) * (DM/_DM)/256; // LAI value following a defoliation
		}// mass in grams
		else{
			res = ((float)DM*(float)sim.METER_SQUARED/10.0f)*(float)LAI/3.0f;
			DM = 0;
			LAI = 0;
		}
		res = factorReal(res);
		return res;
	}
	
	// Returns the value of the grass being trodden on
	public float trample(){ return ((float)DM)*(float)LAI/3.0f; }

	
	public float energyValue(int height){ return ((float)DM)*(float)LAI/3.0f; }


	/** Prints current status */
	public void printState(){
		//System.out.println("Soil Water Capacity (domain:0-200) = "+SWC);
		System.out.println("Available Water (domain:0-250) = "+AW);
		System.out.println("LAI = "+LAI);
		System.out.println("Dry Matter (domain:0-1500) = "+DM);
		System.out.println();
	}

// ------------------------[ Debug ]---------------------------
	
	// Main method for debug and testing
	public static void main(String[] args){

		GrassModel g = new GrassModel(null);

		g.initVars();
		g.setData(50,100,0,0,0);

		//int AW_ = g.getAW();
		//int NI_ = g.getNI();
		//int LAI_ = g.getLAI();
		//int DM_ = g.getDM();
		//AW_ = g.getAW();
		//NI_ = g.getNI();
		//LAI_ = g.getLAI();
		//DM_ = g.getDM();

		//g.setData(AW_, NI_, LAI_, DM_);
		
		//g.prepare();
		//g.printState();
		
		//for(int y = 0; y < 10; y++){
			for(int w=0; w < 52; w++){
				//g.printState();
				for(int d=0; d < 7; d++){

					g.simulate();
					g.printState();

					//if(w==0 && y==0){ System.out.println("Week 1"); g.printState(); }
				}
			}
		//}

	}


}
