/*
 * File:   FieldGate.java
 * Author: Conrad Rider
 * Email:  c.rider@eviltree.co.uk
 * Date:   22-02-05
 */

 
package sim.pasture;

import utils.*;

import java.awt.*;
import java.awt.geom.*;
import sim.*;
import java.util.*;
import java.io.*;


/** These represent the pasture's gates. They may only occur on field boundaries */
public class FieldGate extends Point2D.Double implements FieldItem, Cloneable, Serializable{

	private transient Simulation sim;      // A reference to the simulation
	private transient AnimalContainer ac1; // Neighbour 1
	private transient AnimalContainer ac2; // Neighbour 2
	private           boolean open;        // Whether the gate is open
	private           boolean drawToggle;  // Whether to draw the gate open or closed
	private           Vector2D v;          // The gate's orientation
	
	
	/** Constructor for a new gate */
	public FieldGate(Point2D.Double p, Vector2D v, Simulation sim){
		this.x = p.x;
		this.y = p.y;
		this.v = v;
		open = false;
		drawToggle = false;
		initialise(sim);
	}
	
	/** Initialise the gate after a load */
	public void initialise(Simulation sim){
		this.sim = sim;
		LinkedList fields = sim.state.getFields();
		// Find fields that this gate is connecting
		for(Iterator i = fields.iterator(); i.hasNext();){
			FieldBounds f = (FieldBounds)i.next();
			if(f.ownsGate(this)){
				if(ac1 == null) ac1 = (AnimalContainer)f;
				else{
					ac2 = (AnimalContainer)f;
					return;
				}
			}
		}
		// If we reach here then it means ac2 has not been assigned
		// so the gate must be to the outside world
		ac2 = (AnimalContainer)sim.state;
	}
	
	// Notifies the gate that a field has gone
	public void notifyFieldRemoval(FieldBounds f){
		if(f == ac1) ac1 = null;
		else if(f == ac2) ac2 = null;
	}
/*	
	// Constructs a cloned gate
	public Object clone(){ return new FieldGate(this, sim, v, open); }
	public FieldGate(FieldGate f, Simulation sim, Vector2D v, boolean open){
		this.x = f.x;
		this.y = f.y;
		this.sim = sim;
		this.v = (Vector2D)v.clone();
		this.open = open;
	}
	
	
	public void setSimulation(Simulation sim){ this.sim = sim; }
*/
	
	/** Open or close the gate */
	public void toggle(){
		open = !open;
		FieldBounds f1 = null, f2 = null;
		LinkedList fields = sim.getFieldGeometry().fields;
		for(Iterator i = fields.iterator(); i.hasNext(); ){
			FieldBounds f = (FieldBounds)i.next();
			if(f.ownsGate(this) && f1 != f){
				if(f1 == null) f1 = f;
				else f2 = f;
			}
		}
		sim.notifyGateSwitch();
	}
	public boolean isOpen(){ return open; }
	
	
	/** Returns weather this line will pass between the gate posts */
	public boolean lineFits(Line2D.Double l){
		Vector2D path = new Vector2D((Point2D.Double)l.getP1(), (Point2D.Double)l.getP2()); // Get direction of path
		double angle = Math.acos(Vector2D.dot(v, path)/(v.mag()*path.mag())); // Find angle compared to gate
		double width = Math.sin(angle) * sim.GATE_WIDTH * 0.5; // Proportion of radius is sin of angle
		return l.ptSegDist((Point2D.Double)this) < width; // If line comes close enough to point, then it fits between the posts
	}

	
	public void draw(Graphics g){
		g.setColor(Color.RED);
		drawShapes(g);
	}

	public void drawActive(Graphics g){
		g.setColor(Color.WHITE);
		drawShapes(g);
	}
	
	public void setDrawToggle(){ drawToggle = true;	}
	public void unsetDrawToggle(){ drawToggle = false; }
		
	private void drawShapes(Graphics g){
		if(drawToggle){ g.setColor(Color.PINK); }
		Line2D.Double l = getSegment();
		if(!open) sim.drawScaledLine(g, l.x1, l.y1, l.x2, l.y2);
		sim.fillScaledRect(g, l.x1-sim.GATE_WIDTH/16, l.y1-sim.GATE_WIDTH/16, sim.GATE_WIDTH/8, sim.GATE_WIDTH/8);
		sim.fillScaledRect(g, l.x2-sim.GATE_WIDTH/16, l.y2-sim.GATE_WIDTH/16, sim.GATE_WIDTH/8, sim.GATE_WIDTH/8);
	}
	
	/*public boolean within(Point p1, Point p2){
		Line2D.Double s = new Line2D.Double(p1, p2);
		Line2D.Double g = getSegment();
		//System.out.println("Within is " + (s.contains(g1) && s.contains(g2)) + "  Segment:" + p1 + "-" + p2 + "\n    Does not contain either " + g1 + " or " + g2);
		return contains(s,(Point2D.Double)g.getP1()) && contains(s,(Point2D.Double)g.getP2());			
	}*/
	
	/** Returns weather this point is inside the gates circumferandce */
	public boolean near(Point2D.Double p){ return this.distance(p) <= sim.GATE_WIDTH/2.0; }
	
	/** Returns the nieghhour of the animal container a connected by this gate (if any) */
	public AnimalContainer neighbour(AnimalContainer a){
		if(a == ac1) return ac2;
		else if(a == ac2) return ac1;
		return null;
	}
	
	// Returns the direction moving towards field f
	public Vector2D towards(AnimalContainer f){
		Vector2D d = v.perp();
		if(f.holds(d.project(this, sim.GATE_WIDTH*0.5))) return d;
		d = d.negate();		
		if(f.holds(d.project(this, sim.GATE_WIDTH*0.5))) return d;
		
		// If this is the case then f does not own this gate, return random
		return new Vector2D();
	}
	

	public boolean insideLine(double x1, double y1, double x2, double y2){
		Line2D.Double l = new Line2D.Double(x1,y1,x2,y2);
		Line2D.Double p = getSegment();
		Point2D.Double p1 = (Point2D.Double)p.getP1();
		Point2D.Double p2 = (Point2D.Double)p.getP2();
		return (l.ptSegDist(p1.x, p1.y) <= 0.5) && (l.ptSegDist(p2.x, p2.y) <= 0.5);
	}
		
	public boolean onLine(double x1, double y1, double x2, double y2){
		Line2D.Double l = new Line2D.Double(x1,y1,x2,y2);
		return l.ptSegDist(x, y) <= 0.5;
	}
	
	public static boolean contains(Line2D.Double s, Point2D.Double p){
		return (((p.x < s.x1) && (p.x > s.x2)) || ((p.x > s.x1) && (p.x < s.x2))
		     || ((p.y < s.y1) && (p.y > s.y2)) || ((p.y > s.y1) && (p.y < s.y2)));
	}
	
	public Line2D.Double getSegment(){
		Vector2D nv = v.negate();
		Point2D.Double p1 = v.project(this, sim.GATE_WIDTH/2);
		Point2D.Double p2 = nv.project(this, sim.GATE_WIDTH/2);
		return new Line2D.Double(p1, p2);
	}
	
	public String toString(){ return "FieldGate["+x+","+y+"]"; }

	public boolean overlap(FieldGate g){
		Point2D.Double p = new Point2D.Double(x,y);
		if(p.distance(new Point2D.Double(g.x, g.y)) <= sim.GATE_WIDTH) return true;
		return false;
	}
	
	public boolean hasField(LinkedList fields){
		for(Iterator i = fields.iterator(); i.hasNext();){
			FieldBounds f = (FieldBounds)i.next();
			if(f.ownsGate(this)) return true;
		}
		return false;
	}
	
	private void writeObject(java.io.ObjectOutputStream out) throws IOException{
		out.writeDouble(x);
		out.writeDouble(y);
		out.writeBoolean(open);
		out.writeObject(v);
	}
	private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException{
		x = in.readDouble();
		y = in.readDouble();
		open = in.readBoolean();
		v = (Vector2D)in.readObject();
	}

	//public boolean equals(Object o){ return o instanceof FieldGate && id = ((FieldGate)o).id(); }
}
