/*
 * File:   FileManager.java
 * Author: Conrad Rider
 * Email:  c.rider@eviltree.co.uk
 * Date:   22-02-05
 */

 
 package utils;
 
 import sim.*;
 import gui.*;
 
 import java.io.*;
 import java.awt.*;
 import java.awt.image.*;
 import javax.imageio.*;
 
/**
 * This class organises the simulation data so that it can
 * be saved in one file. The organisation is as follows:
 * 
 *  ------------------------------------------------------------------------
 * | Base Simulation Address   |  History Count   |   Last Block Address    |  <-- Main header
 * |------------------------------------------------------------------------|
 * | 2nd History Address  |  3rd History Address | ... | Next Block Address |  <-- Block header
 * |------------------------------------------------------------------------|
 * | 1st State History                                                      |  
 * | 2nd State History                                                      |  <-- State history data
 * | ...                                                                    |
 * | 10th State History                                                     |
 * |------------------------------------------------------------------------|
 * | 12th History Address | 13th History Address | ... | Next Block Address |  <-- Block header
 * |------------------------------------------------------------------------|
 * | 12th State History                                                     |
 * | 13th State History                                                     |  <-- State history data
 * | ...                                                                    |  
 * |------------------------------------------------------------------------|
 * | Simulation                                                             |  <-- Simulation data
 * |                                                                        |
 *  ------------------------------------------------------------------------
 * 
 */
 public class FileManager extends RandomAccessFile{
	
 	private static final int BLOCK_SIZE = 2;
	private static final int LONG_SIZE = 8;
	private static final int INT_SIZE = 4;
 	private static final int HEADER_SIZE = INT_SIZE + 2*LONG_SIZE;
	
 
	private Simulation sim;
 
	private long sAddr;
	private int hCount;
	private long lbAddr;
	
	private int lock = 0;
	private long timeout;
		 
	public FileManager(File file, Simulation sim) throws FileNotFoundException{
		super(file, "rw");
		this.sim = sim;
		
		sAddr = -1;
		hCount = -1;
		lbAddr = -1;
		
		lock = 0;
		
	}
	
	
	/*-----------------[ Top level public functions ]------------------*/
	
	public boolean isSaving(){ return lock != 0; }
	
	
	/** Creates a new simulation */
	public void createSimulation(FrozenState state){
		assertControl(1);
		try{
			// Write the address of the simulation object
			// this comes after the block header, which comes after
			// this header.
			setSimAddr(HEADER_SIZE+(BLOCK_SIZE*LONG_SIZE));
			// Write the history count
			setHistoryCount(0);
			// Write the last block address, this is the next
			// thing after this header
			setLastBlockAddr(HEADER_SIZE);
	
			saveState(state);
			saveSim();
				
		}catch(Exception e){
			//Popup.error("ERROR: Create failed!");
			System.err.println("ERROR: Couln't write simulation file. [" + e +"]");
			e.printStackTrace();
		}
		releaseControl(1);
	}	

	
	/** Loads the simulation */
	public FrozenState loadSimulation(){
		assertControl(2);
		FrozenState state = null;
		try{
			state = loadState(historyCount()-1);
			loadSim();
		}catch(Exception e){
			System.err.println("ERROR: Couln't read simulation data. [" + e +"]");
			e.printStackTrace();
		}
		releaseControl(2);
		return state;
	}
	
		
	/** Saves the simulation */
	public void saveSimulation(final FrozenState last){
		assertControl(3);
		Thread t = new Thread(){
			public void run(){
				int hc = historyCount();
				try{
					saveState(last);
					saveSim();
				}catch(Exception e){
					//Popup.error("ERROR: Simulation Save Failed!");
					System.err.println("ERROR: Couln't write simulation data. [" + e +"]");
					e.printStackTrace();
				}
				releaseControl(3);
			}
		};
		t.setName("Simulation Save");
		t.setPriority(Thread.MIN_PRIORITY);
		t.start();
	}

	
	
	/** Copies the simulation to a new file */
	public FileManager copySimulation(File newFile){
		assertControl(4);
		FileManager newManager = null;
		try{
			saveSim();
			copySim(newFile);
			newManager = new FileManager(newFile, sim);
		}catch(Exception e){
			//Popup.error("ERROR: Simulation Save Failed!");
			System.err.println("ERROR: Couln't copy simulation data. [" + e +"]");
			e.printStackTrace();
		}
		releaseControl(4);
		return newManager;
	}
	
	public void export(final File outFile){
		assertControl(5);
		Thread t = new Thread(){
			public void run(){
				try{
					Dimension d = sim.getPastureSize();
					BufferedImage image = new BufferedImage(d.width, d.height, BufferedImage.TYPE_INT_ARGB);
					Graphics g = image.createGraphics();
					sim.drawPasture(g);
					ImageIO.write(image, "png", outFile);
					g.dispose(); 
				}catch(Exception e){
					//Popup.error("ERROR: Simulation Save Failed!");
					System.err.println("ERROR: Couln't write simulation data. [" + e +"]");
					e.printStackTrace();
				}
				releaseControl(5);
			}
		};
		t.setName("Export Simulation Graphics");
		t.start();
	}
		
	
	/** Loads a state history object */
	public FrozenState loadHistory(int n){
		assertControl(6);
		FrozenState state = null;
		try{
			state = loadState(n);
		}catch(Exception e){
			//Popup.error("ERROR: History Load Failed!");
			System.err.println("ERROR: Couln't read simulation data. [" + e +"]");
			e.printStackTrace();
		}
		releaseControl(6);
		return state;
	}
	
			
	public void close(){
		final RandomAccessFile f = this;
		new Thread(){
			public void run(){
				assertControl(10);
				try{
					f.close();
				}catch(IOException e){ e.printStackTrace(); }
			}
		}.start();
	}
	
	
	
	/*----------------[ Non-threaded operatoins ]------------*/
	
	
	/** Loads the simulation */
	private void loadSim() throws Exception{
		seek(simAddr());
		int len = readInt();
		byte[] buf = new byte[len];
		ByteArrayInputStream input = new ByteArrayInputStream(buf);
		read(buf, 0, len);
		ObjectInputStream objIn = new ObjectInputStream(input);
		sim.loadSimulation(objIn);
		objIn.close();
		input = null;
		objIn = null;
		Runtime.getRuntime().gc();
	}

			
	/** Saves the simulation */
	private void saveSim() throws Exception{
		ByteArrayOutputStream output = new ByteArrayOutputStream();
		ObjectOutputStream objOut = new ObjectOutputStream(output);
		sim.saveSimulation(objOut);
		objOut.close();
		seek(simAddr());
		writeInt(output.size());
		write(output.toByteArray());
		output.close();
	}
	
	
	
	
	/** Loads a state history object */
	private FrozenState loadState(int n) throws Exception{
		//while(!sim.enoughMemory()){ Runtime.getRuntime().gc(); Thread.currentThread().yield(); }

		FrozenState state = null;
		long addr = frozenStateAddr(n); seek(addr);
		int len = readInt();
		byte[] buf = new byte[len];
		ByteArrayInputStream input = new ByteArrayInputStream(buf);
		read(buf, 0, len);
		ObjectInputStream objIn = new ObjectInputStream(input);
		state = (FrozenState)objIn.readObject();
		objIn.close();
		return state;
	}
	
	
			
	/** Saves a state history object */
	private void saveState(FrozenState state) throws Exception{
		ByteArrayOutputStream output = new ByteArrayOutputStream();
		ObjectOutputStream objOut = new ObjectOutputStream(output);
		objOut.writeObject(state);
		objOut.close();
		long addr = nextFrozenStateAddr(); seek(addr);
		writeInt(output.size());
		write(output.toByteArray());
		writeFrozenStateAddr();
		setHistoryCount(historyCount()+1);
		output = null;
		objOut = null;
		Runtime.getRuntime().gc();
	}
	
	
	/** Copies the simulation to a new file */
	private void copySim(File newFile) throws Exception{
		InputStream input = new FileInputStream(sim.getSimulationFile());
		OutputStream output = new FileOutputStream(newFile);
		byte[] buffer = new byte[1024]; int len;
		while((len = input.read(buffer)) > 0){
			output.write(buffer, 0, len);
		} output.close();
	}

	
	
	/*-------------------[ Seeker functions ]----------------*/
	

	
	public void writeFrozenStateAddr() throws Exception{
		long endAddr = getFilePointer();
		int offset = historyCount() % BLOCK_SIZE;
		seek(lastBlockAddr() + offset*LONG_SIZE);
		writeLong(endAddr);
		
		setSimAddr(endAddr);
		if(offset == BLOCK_SIZE-1){
			// Incrament the last block address
			setLastBlockAddr(endAddr);
		}
	}
	
		
	/*-------------------[ Addressing Functions ]----------------*/
	
	
	private long frozenStateAddr(int n) throws Exception{
		
		int block = (int)(n / BLOCK_SIZE);
		int offset = n % BLOCK_SIZE;
		
		long addr = 0;
		
		if(block == 0){
			if(offset == 0){
				seek(HEADER_SIZE + BLOCK_SIZE*LONG_SIZE);
				addr = HEADER_SIZE + BLOCK_SIZE*LONG_SIZE;
			}else{
				seek(HEADER_SIZE + (offset - 1)*LONG_SIZE);
				addr = readLong();
			}
		}else{
			seek(HEADER_SIZE + (BLOCK_SIZE-1)*LONG_SIZE);
			long nextBlockAddr = readLong();
			for(int i = 1; i < block; i++){
				seek(nextBlockAddr+(BLOCK_SIZE-1)*LONG_SIZE);
				nextBlockAddr = readLong();
			}
			
			// Search for the offset within the block
			if(offset == 0){
				addr = (nextBlockAddr + BLOCK_SIZE*LONG_SIZE);
			}else{
				seek(nextBlockAddr + (offset-1)*LONG_SIZE);
				addr = readLong();
			}
		}
		return addr;
	}

	
	private long nextFrozenStateAddr() throws Exception{

		long addr;
		
		
		int offset = historyCount() % BLOCK_SIZE;
		// If the last saved history is the last in
		// that block we will need to leave room for
		// the next block header
		if(offset == 0){
			addr = lastBlockAddr() + BLOCK_SIZE*LONG_SIZE;
		}else{
			seek(lastBlockAddr() + (offset-1)*LONG_SIZE);
			addr = readLong();
		}
		return addr;
	}

		
	private long simAddr() throws Exception{
		if(sAddr == -1){
			seek(0);
			sAddr = readLong();
		}
		return sAddr;
	}private void setSimAddr(long addr) throws Exception{
		seek(0);
		writeLong(addr);
		sAddr = addr;
	}
	
	public int historyCount(){
		try{
			if(hCount == -1){
				seek(LONG_SIZE);
				hCount = readInt();
			}
		}catch(Exception e){
			e.printStackTrace();
			System.err.println("ERROR: Failed to read history count.");
		}
		
		return hCount;
	}private void setHistoryCount(int count) throws Exception{
		seek(LONG_SIZE);
		writeInt(count);
		hCount = count;
	}
	
	private long lastBlockAddr() throws Exception{
		if(lbAddr == -1){
			seek(LONG_SIZE+INT_SIZE);
			lbAddr = readLong();
		}
		return lbAddr;
	}private void setLastBlockAddr(long addr) throws Exception{
		seek(LONG_SIZE+INT_SIZE);
		writeLong(addr);
		lbAddr = addr;
	}

	
 	/*----------------[ Misc functions ]-------------*/
	
	
	private void assertControl(int n){
		while(lock != 0);
		lock = n;
		try{
			Thread.currentThread().sleep(500);
		}
		catch(InterruptedException e){e.printStackTrace();}
	}	
	private void releaseControl(int n){ if(lock == n) lock = 0; }
	
 }
